#------------------------------------------------------------------------------
# pycparser: c_parser.py
#
# CParser class: Parser and AST builder for the C language
#
# Eli Bendersky [https://eli.thegreenplace.net/]
# License: BSD
#------------------------------------------------------------------------------
import re

from .ply import yacc

from . import c_ast
from .c_lexer import CLexer
from .plyparser import PLYParser, Coord, ParseError, parameterized, template
from .ast_transforms import fix_switch_cases


@template
class CParser(PLYParser):
    def __init__(
            self,
            lex_optimize=True,
            lexer=CLexer,
            lextab='pycparser.lextab',
            yacc_optimize=True,
            yacctab='pycparser.yacctab',
            yacc_debug=False,
            taboutputdir=''):
        """ Create a new CParser.

            Some arguments for controlling the debug/optimization
            level of the parser are provided. The defaults are
            tuned for release/performance mode.
            The simple rules for using them are:
            *) When tweaking CParser/CLexer, set these to False
            *) When releasing a stable parser, set to True

            lex_optimize:
                Set to False when you're modifying the lexer.
                Otherwise, changes in the lexer won't be used, if
                some lextab.py file exists.
                When releasing with a stable lexer, set to True
                to save the re-generation of the lexer table on
                each run.

            lexer:
                Set this parameter to define the lexer to use if
                you're not using the default CLexer.

            lextab:
                Points to the lex table that's used for optimized
                mode. Only if you're modifying the lexer and want
                some tests to avoid re-generating the table, make
                this point to a local lex table file (that's been
                earlier generated with lex_optimize=True)

            yacc_optimize:
                Set to False when you're modifying the parser.
                Otherwise, changes in the parser won't be used, if
                some parsetab.py file exists.
                When releasing with a stable parser, set to True
                to save the re-generation of the parser table on
                each run.

            yacctab:
                Points to the yacc table that's used for optimized
                mode. Only if you're modifying the parser, make
                this point to a local yacc table file

            yacc_debug:
                Generate a parser.out file that explains how yacc
                built the parsing table from the grammar.

            taboutputdir:
                Set this parameter to control the location of generated
                lextab and yacctab files.
        """
        self.clex = lexer(
            error_func=self._lex_error_func,
            on_lbrace_func=self._lex_on_lbrace_func,
            on_rbrace_func=self._lex_on_rbrace_func,
            type_lookup_func=self._lex_type_lookup_func)

        self.clex.build(
            optimize=lex_optimize,
            lextab=lextab,
            outputdir=taboutputdir)
        self.tokens = self.clex.tokens

        rules_with_opt = [
            'abstract_declarator',
            'assignment_expression',
            'declaration_list',
            'declaration_specifiers_no_type',
            'designation',
            'expression',
            'identifier_list',
            'init_declarator_list',
            'id_init_declarator_list',
            'initializer_list',
            'parameter_type_list',
            'block_item_list',
            'type_qualifier_list',
            'struct_declarator_list'
        ]

        for rule in rules_with_opt:
            self._create_opt_rule(rule)

        self.cparser = yacc.yacc(
            module=self,
            start='translation_unit_or_empty',
            debug=yacc_debug,
            optimize=yacc_optimize,
            tabmodule=yacctab,
            outputdir=taboutputdir)

        # Stack of scopes for keeping track of symbols. _scope_stack[-1] is
        # the current (topmost) scope. Each scope is a dictionary that
        # specifies whether a name is a type. If _scope_stack[n][name] is
        # True, 'name' is currently a type in the scope. If it's False,
        # 'name' is used in the scope but not as a type (for instance, if we
        # saw: int name;
        # If 'name' is not a key in _scope_stack[n] then 'name' was not defined
        # in this scope at all.
        self._scope_stack = [dict()]

        # Keeps track of the last token given to yacc (the lookahead token)
        self._last_yielded_token = None

    def parse(self, text, filename='', debuglevel=0):
        """ Parses C code and returns an AST.

            text:
                A string containing the C source code

            filename:
                Name of the file being parsed (for meaningful
                error messages)

            debuglevel:
                Debug level to yacc
        """
        self.clex.filename = filename
        self.clex.reset_lineno()
        self._scope_stack = [dict()]
        self._last_yielded_token = None
        return self.cparser.parse(
                input=text,
                lexer=self.clex,
                debug=debuglevel)

    ######################--   PRIVATE   --######################

    def _push_scope(self):
        self._scope_stack.append(dict())

    def _pop_scope(self):
        assert len(self._scope_stack) > 1
        self._scope_stack.pop()

    def _add_typedef_name(self, name, coord):
        """ Add a new typedef name (ie a TYPEID) to the current scope
        """
        if not self._scope_stack[-1].get(name, True):
            self._parse_error(
                "Typedef %r previously declared as non-typedef "
                "in this scope" % name, coord)
        self._scope_stack[-1][name] = True

    def _add_identifier(self, name, coord):
        """ Add a new object, function, or enum member name (ie an ID) to the
            current scope
        """
        if self._scope_stack[-1].get(name, False):
            self._parse_error(
                "Non-typedef %r previously declared as typedef "
                "in this scope" % name, coord)
        self._scope_stack[-1][name] = False

    def _is_type_in_scope(self, name):
        """ Is *name* a typedef-name in the current scope?
        """
        for scope in reversed(self._scope_stack):
            # If name is an identifier in this scope it shadows typedefs in
            # higher scopes.
            in_scope = scope.get(name)
            if in_scope is not None: return in_scope
        return False

    def _lex_error_func(self, msg, line, column):
        self._parse_error(msg, self._coord(line, column))

    def _lex_on_lbrace_func(self):
        self._push_scope()

    def _lex_on_rbrace_func(self):
        self._pop_scope()

    def _lex_type_lookup_func(self, name):
        """ Looks up types that were previously defined with
            typedef.
            Passed to the lexer for recognizing identifiers that
            are types.
        """
        is_type = self._is_type_in_scope(name)
        return is_type

    def _get_yacc_lookahead_token(self):
        """ We need access to yacc's lookahead token in certain cases.
            This is the last token yacc requested from the lexer, so we
            ask the lexer.
        """
        return self.clex.last_token

    # To understand what's going on here, read sections A.8.5 and
    # A.8.6 of K&R2 very carefully.
    #
    # A C type consists of a basic type declaration, with a list
    # of modifiers. For example:
    #
    # int *c[5];
    #
    # The basic declaration here is 'int c', and the pointer and
    # the array are the modifiers.
    #
    # Basic declarations are represented by TypeDecl (from module c_ast) and the
    # modifiers are FuncDecl, PtrDecl and ArrayDecl.
    #
    # The standard states that whenever a new modifier is parsed, it should be
    # added to the end of the list of modifiers. For example:
    #
    # K&R2 A.8.6.2: Array Declarators
    #
    # In a declaration T D where D has the form
    #   D1 [constant-expression-opt]
    # and the type of the identifier in the declaration T D1 is
    # "type-modifier T", the type of the
    # identifier of D is "type-modifier array of T"
    #
    # This is what this method does. The declarator it receives
    # can be a list of declarators ending with TypeDecl. It
    # tacks the modifier to the end of this list, just before
    # the TypeDecl.
    #
    # Additionally, the modifier may be a list itself. This is
    # useful for pointers, that can come as a chain from the rule
    # p_pointer. In this case, the whole modifier list is spliced
    # into the new location.
    def _type_modify_decl(self, decl, modifier):
        """ Tacks a type modifier on a declarator, and returns
            the modified declarator.

            Note: the declarator and modifier may be modified
        """
        #~ print '****'
        #~ decl.show(offset=3)
        #~ modifier.show(offset=3)
        #~ print '****'

        modifier_head = modifier
        modifier_tail = modifier

        # The modifier may be a nested list. Reach its tail.
        #
        while modifier_tail.type:
            modifier_tail = modifier_tail.type

        # If the decl is a basic type, just tack the modifier onto
        # it
        #
        if isinstance(decl, c_ast.TypeDecl):
            modifier_tail.type = decl
            return modifier
        else:
            # Otherwise, the decl is a list of modifiers. Reach
            # its tail and splice the modifier onto the tail,
            # pointing to the underlying basic type.
            #
            decl_tail = decl

            while not isinstance(decl_tail.type, c_ast.TypeDecl):
                decl_tail = decl_tail.type

            modifier_tail.type = decl_tail.type
            decl_tail.type = modifier_head
            return decl

    # Due to the order in which declarators are constructed,
    # they have to be fixed in order to look like a normal AST.
    #
    # When a declaration arrives from syntax construction, it has
    # these problems:
    # * The innermost TypeDecl has no type (because the basic
    #   type is only known at the uppermost declaration level)
    # * The declaration has no variable name, since that is saved
    #   in the innermost TypeDecl
    # * The typename of the declaration is a list of type
    #   specifiers, and not a node. Here, basic identifier types
    #   should be separated from more complex types like enums
    #   and structs.
    #
    # This method fixes these problems.
    #
    def _fix_decl_name_type(self, decl, typename):
        """ Fixes a declaration. Modifies decl.
        """
        # Reach the underlying basic type
        #
        type = decl
        while not isinstance(type, c_ast.TypeDecl):
            type = type.type

        decl.name = type.declname
        type.quals = decl.quals

        # The typename is a list of types. If any type in this
        # list isn't an IdentifierType, it must be the only
        # type in the list (it's illegal to declare "int enum ..")
        # If all the types are basic, they're collected in the
        # IdentifierType holder.
        #
        for tn in typename:
            if not isinstance(tn, c_ast.IdentifierType):
                if len(typename) > 1:
                    self._parse_error(
                        "Invalid multiple types specified", tn.coord)
                else:
                    type.type = tn
                    return decl

        if not typename:
            # Functions default to returning int
            #
            if not isinstance(decl.type, c_ast.FuncDecl):
                self._parse_error(
                        "Missing type in declaration", decl.coord)
            type.type = c_ast.IdentifierType(
                    ['int'],
                    coord=decl.coord)
        else:
            # At this point, we know that typename is a list of IdentifierType
            # nodes. Concatenate all the names into a single list.
            #
            type.type = c_ast.IdentifierType(
                [name for id in typename for name in id.names],
                coord=typename[0].coord)
        return decl

    def _add_declaration_specifier(self, declspec, newspec, kind, append=False):
        """ Declaration specifiers are represented by a dictionary
            with the entries:
            * qual: a list of type qualifiers
            * storage: a list of storage type qualifiers
            * type: a list of type specifiers
            * function: a list of function specifiers

            This method is given a declaration specifier, and a
            new specifier of a given kind.
            If `append` is True, the new specifier is added to the end of
            the specifiers list, otherwise it's added at the beginning.
            Returns the declaration specifier, with the new
            specifier incorporated.
        """
        spec = declspec or dict(qual=[], storage=[], type=[], function=[])

        if append:
            spec[kind].append(newspec)
        else:
            spec[kind].insert(0, newspec)

        return spec

    def _build_declarations(self, spec, decls, typedef_namespace=False):
        """ Builds a list of declarations all sharing the given specifiers.
            If typedef_namespace is true, each declared name is added
            to the "typedef namespace", which also includes objects,
            functions, and enum constants.
        """
        is_typedef = 'typedef' in spec['storage']
        declarations = []

        # Bit-fields are allowed to be unnamed.
        #
        if decls[0].get('bitsize') is not None:
            pass

        # When redeclaring typedef names as identifiers in inner scopes, a
        # problem can occur where the identifier gets grouped into
        # spec['type'], leaving decl as None.  This can only occur for the
        # first declarator.
        #
        elif decls[0]['decl'] is None:
            if len(spec['type']) < 2 or len(spec['type'][-1].names) != 1 or \
                    not self._is_type_in_scope(spec['type'][-1].names[0]):
                coord = '?'
                for t in spec['type']:
                    if hasattr(t, 'coord'):
                        coord = t.coord
                        break
                self._parse_error('Invalid declaration', coord)

            # Make this look as if it came from "direct_declarator:ID"
            decls[0]['decl'] = c_ast.TypeDecl(
                declname=spec['type'][-1].names[0],
                type=None,
                quals=None,
                coord=spec['type'][-1].coord)
            # Remove the "new" type's name from the end of spec['type']
            del spec['type'][-1]

        # A similar problem can occur where the declaration ends up looking
        # like an abstract declarator.  Give it a name if this is the case.
        #
        elif not isinstance(decls[0]['decl'],
                (c_ast.Struct, c_ast.Union, c_ast.IdentifierType)):
            decls_0_tail = decls[0]['decl']
            while not isinstance(decls_0_tail, c_ast.TypeDecl):
                decls_0_tail = decls_0_tail.type
            if decls_0_tail.declname is None:
                decls_0_tail.declname = spec['type'][-1].names[0]
                del spec['type'][-1]

        for decl in decls:
            assert decl['decl'] is not None
            if is_typedef:
                declaration = c_ast.Typedef(
                    name=None,
                    quals=spec['qual'],
                    storage=spec['storage'],
                    type=decl['decl'],
                    coord=decl['decl'].coord)
            else:
                declaration = c_ast.Decl(
                    name=None,
                    quals=spec['qual'],
                    storage=spec['storage'],
                    funcspec=spec['function'],
                    type=decl['decl'],
                    init=decl.get('init'),
                    bitsize=decl.get('bitsize'),
                    coord=decl['decl'].coord)

            if isinstance(declaration.type,
                    (c_ast.Struct, c_ast.Union, c_ast.IdentifierType)):
                fixed_decl = declaration
            else:
                fixed_decl = self._fix_decl_name_type(declaration, spec['type'])

            # Add the type name defined by typedef to a
            # symbol table (for usage in the lexer)
            #
            if typedef_namespace:
                if is_typedef:
                    self._add_typedef_name(fixed_decl.name, fixed_decl.coord)
                else:
                    self._add_identifier(fixed_decl.name, fixed_decl.coord)

            declarations.append(fixed_decl)

        return declarations

    def _build_function_definition(self, spec, decl, param_decls, body):
        """ Builds a function definition.
        """
        assert 'typedef' not in spec['storage']

        declaration = self._build_declarations(
            spec=spec,
            decls=[dict(decl=decl, init=None)],
            typedef_namespace=True)[0]

        return c_ast.FuncDef(
            decl=declaration,
            param_decls=param_decls,
            body=body,
            coord=decl.coord)

    def _select_struct_union_class(self, token):
        """ Given a token (either STRUCT or UNION), selects the
            appropriate AST class.
        """
        if token == 'struct':
            return c_ast.Struct
        else:
            return c_ast.Union

    ##
    ## Precedence and associativity of operators
    ##
    precedence = (
        ('left', 'LOR'),
        ('left', 'LAND'),
        ('left', 'OR'),
        ('left', 'XOR'),
        ('left', 'AND'),
        ('left', 'EQ', 'NE'),
        ('left', 'GT', 'GE', 'LT', 'LE'),
        ('left', 'RSHIFT', 'LSHIFT'),
        ('left', 'PLUS', 'MINUS'),
        ('left', 'TIMES', 'DIVIDE', 'MOD')
    )

    ##
    ## Grammar productions
    ## Implementation of the BNF defined in K&R2 A.13
    ##

    # Wrapper around a translation unit, to allow for empty input.
    # Not strictly part of the C99 Grammar, but useful in practice.
    #
    def p_translation_unit_or_empty(self, p):
        """ translation_unit_or_empty   : translation_unit
                                        | empty
        """
        if p[1] is None:
            p[0] = c_ast.FileAST([])
        else:
            p[0] = c_ast.FileAST(p[1])

    def p_translation_unit_1(self, p):
        """ translation_unit    : external_declaration
        """
        # Note: external_declaration is already a list
        #
        p[0] = p[1]

    def p_translation_unit_2(self, p):
        """ translation_unit    : translation_unit external_declaration
        """
        if p[2] is not None:
            p[1].extend(p[2])
        p[0] = p[1]

    # Declarations always come as lists (because they can be
    # several in one line), so we wrap the function definition
    # into a list as well, to make the return value of
    # external_declaration homogenous.
    #
    def p_external_declaration_1(self, p):
        """ external_declaration    : function_definition
        """
        p[0] = [p[1]]

    def p_external_declaration_2(self, p):
        """ external_declaration    : declaration
        """
        p[0] = p[1]

    def p_external_declaration_3(self, p):
        """ external_declaration    : pp_directive
                                    | pppragma_directive
        """
        p[0] = [p[1]]

    def p_external_declaration_4(self, p):
        """ external_declaration    : SEMI
        """
        p[0] = None

    def p_pp_directive(self, p):
        """ pp_directive  : PPHASH
        """
        self._parse_error('Directives not supported yet',
                          self._token_coord(p, 1))

    def p_pppragma_directive(self, p):
        """ pppragma_directive      : PPPRAGMA
                                    | PPPRAGMA PPPRAGMASTR
        """
        if len(p) == 3:
            p[0] = c_ast.Pragma(p[2], self._token_coord(p, 2))
        else:
            p[0] = c_ast.Pragma("", self._token_coord(p, 1))

    # In function definitions, the declarator can be followed by
    # a declaration list, for old "K&R style" function definitios.
    #
    def p_function_definition_1(self, p):
        """ function_definition : id_declarator declaration_list_opt compound_statement
        """
        # no declaration specifiers - 'int' becomes the default type
        spec = dict(
            qual=[],
            storage=[],
            type=[c_ast.IdentifierType(['int'],
                                       coord=self._token_coord(p, 1))],
            function=[])

        p[0] = self._build_function_definition(
            spec=spec,
            decl=p[1],
            param_decls=p[2],
            body=p[3])

    def p_function_definition_2(self, p):
        """ function_definition : declaration_specifiers id_declarator declaration_list_opt compound_statement
        """
        spec = p[1]

        p[0] = self._build_function_definition(
            spec=spec,
            decl=p[2],
            param_decls=p[3],
            body=p[4])

    def p_statement(self, p):
        """ statement   : labeled_statement
                        | expression_statement
                        | compound_statement
                        | selection_statement
                        | iteration_statement
                        | jump_statement
                        | pppragma_directive
        """
        p[0] = p[1]

    # A pragma is generally considered a decorator rather than an actual statement.
    # Still, for the purposes of analyzing an abstract syntax tree of C code,
    # pragma's should not be ignored and were previously treated as a statement.
    # This presents a problem for constructs that take a statement such as labeled_statements,
    # selection_statements, and iteration_statements, causing a misleading structure
    # in the AST. For example, consider the following C code.
    #
    #   for (int i = 0; i < 3; i++)
    #       #pragma omp critical
    #       sum += 1;
    #
    # This code will compile and execute "sum += 1;" as the body of the for loop.
    # Previous implementations of PyCParser would render the AST for this
    # block of code as follows:
    #
    #   For:
    #     DeclList:
    #       Decl: i, [], [], []
    #         TypeDecl: i, []
    #           IdentifierType: ['int']
    #         Constant: int, 0
    #     BinaryOp: <
    #       ID: i
    #       Constant: int, 3
    #     UnaryOp: p++
    #       ID: i
    #     Pragma: omp critical
    #   Assignment: +=
    #     ID: sum
    #     Constant: int, 1
    #
    # This AST misleadingly takes the Pragma as the body of the loop and the
    # assignment then becomes a sibling of the loop.
    #
    # To solve edge cases like these, the pragmacomp_or_statement rule groups
    # a pragma and its following statement (which would otherwise be orphaned)
    # using a compound block, effectively turning the above code into:
    #
    #   for (int i = 0; i < 3; i++) {
    #       #pragma omp critical
    #       sum += 1;
    #   }
    def p_pragmacomp_or_statement(self, p):
        """ pragmacomp_or_statement     : pppragma_directive statement
                                        | statement
        """
        if isinstance(p[1], c_ast.Pragma) and len(p) == 3:
            p[0] = c_ast.Compound(
                block_items=[p[1], p[2]],
                coord=self._token_coord(p, 1))
        else:
            p[0] = p[1]

    # In C, declarations can come several in a line:
    #   int x, *px, romulo = 5;
    #
    # However, for the AST, we will split them to separate Decl
    # nodes.
    #
    # This rule splits its declarations and always returns a list
    # of Decl nodes, even if it's one element long.
    #
    def p_decl_body(self, p):
        """ decl_body : declaration_specifiers init_declarator_list_opt
                      | declaration_specifiers_no_type id_init_declarator_list_opt
        """
        spec = p[1]

        # p[2] (init_declarator_list_opt) is either a list or None
        #
        if p[2] is None:
            # By the standard, you must have at least one declarator unless
            # declaring a structure tag, a union tag, or the members of an
            # enumeration.
            #
            ty = spec['type']
            s_u_or_e = (c_ast.Struct, c_ast.Union, c_ast.Enum)
            if len(ty) == 1 and isinstance(ty[0], s_u_or_e):
                decls = [c_ast.Decl(
                    name=None,
                    quals=spec['qual'],
                    storage=spec['storage'],
                    funcspec=spec['function'],
                    type=ty[0],
                    init=None,
                    bitsize=None,
                    coord=ty[0].coord)]

            # However, this case can also occur on redeclared identifiers in
            # an inner scope.  The trouble is that the redeclared type's name
            # gets grouped into declaration_specifiers; _build_declarations
            # compensates for this.
            #
            else:
                decls = self._build_declarations(
                    spec=spec,
                    decls=[dict(decl=None, init=None)],
                    typedef_namespace=True)

        else:
            decls = self._build_declarations(
                spec=spec,
                decls=p[2],
                typedef_namespace=True)

        p[0] = decls

    # The declaration has been split to a decl_body sub-rule and
    # SEMI, because having them in a single rule created a problem
    # for defining typedefs.
    #
    # If a typedef line was directly followed by a line using the
    # type defined with the typedef, the type would not be
    # recognized. This is because to reduce the declaration rule,
    # the parser's lookahead asked for the token after SEMI, which
    # was the type from the next line, and the lexer had no chance
    # to see the updated type symbol table.
    #
    # Splitting solves this problem, because after seeing SEMI,
    # the parser reduces decl_body, which actually adds the new
    # type into the table to be seen by the lexer before the next
    # line is reached.
    def p_declaration(self, p):
        """ declaration : decl_body SEMI
        """
        p[0] = p[1]

    # Since each declaration is a list of declarations, this
    # rule will combine all the declarations and return a single
    # list
    #
    def p_declaration_list(self, p):
        """ declaration_list    : declaration
                                | declaration_list declaration
        """
        p[0] = p[1] if len(p) == 2 else p[1] + p[2]

    # To know when declaration-specifiers end and declarators begin,
    # we require declaration-specifiers to have at least one
    # type-specifier, and disallow typedef-names after we've seen any
    # type-specifier. These are both required by the spec.
    #
    def p_declaration_specifiers_no_type_1(self, p):
        """ declaration_specifiers_no_type  : type_qualifier declaration_specifiers_no_type_opt
        """
        p[0] = self._add_declaration_specifier(p[2], p[1], 'qual')

    def p_declaration_specifiers_no_type_2(self, p):
        """ declaration_specifiers_no_type  : storage_class_specifier declaration_specifiers_no_type_opt
        """
        p[0] = self._add_declaration_specifier(p[2], p[1], 'storage')

    def p_declaration_specifiers_no_type_3(self, p):
        """ declaration_specifiers_no_type  : function_specifier declaration_specifiers_no_type_opt
        """
        p[0] = self._add_declaration_specifier(p[2], p[1], 'function')


    def p_declaration_specifiers_1(self, p):
        """ declaration_specifiers  : declaration_specifiers type_qualifier
        """
        p[0] = self._add_declaration_specifier(p[1], p[2], 'qual', append=True)

    def p_declaration_specifiers_2(self, p):
        """ declaration_specifiers  : declaration_specifiers storage_class_specifier
        """
        p[0] = self._add_declaration_specifier(p[1], p[2], 'storage', append=True)

    def p_declaration_specifiers_3(self, p):
        """ declaration_specifiers  : declaration_specifiers function_specifier
        """
        p[0] = self._add_declaration_specifier(p[1], p[2], 'function', append=True)

    def p_declaration_specifiers_4(self, p):
        """ declaration_specifiers  : declaration_specifiers type_specifier_no_typeid
        """
        p[0] = self._add_declaration_specifier(p[1], p[2], 'type', append=True)

    def p_declaration_specifiers_5(self, p):
        """ declaration_specifiers  : type_specifier
        """
        p[0] = self._add_declaration_specifier(None, p[1], 'type')

    def p_declaration_specifiers_6(self, p):
        """ declaration_specifiers  : declaration_specifiers_no_type type_specifier
        """
        p[0] = self._add_declaration_specifier(p[1], p[2], 'type', append=True)


    def p_storage_class_specifier(self, p):
        """ storage_class_specifier : AUTO
                                    | REGISTER
                                    | STATIC
                                    | EXTERN
                                    | TYPEDEF
        """
        p[0] = p[1]

    def p_function_specifier(self, p):
        """ function_specifier  : INLINE
        """
        p[0] = p[1]

    def p_type_specifier_no_typeid(self, p):
        """ type_specifier_no_typeid  : VOID
                                      | _BOOL
                                      | CHAR
                                      | SHORT
                                      | INT
                                      | LONG
                                      | FLOAT
                                      | DOUBLE
                                      | _COMPLEX
                                      | SIGNED
                                      | UNSIGNED
                                      | __INT128
        """
        p[0] = c_ast.IdentifierType([p[1]], coord=self._token_coord(p, 1))

    def p_type_specifier(self, p):
        """ type_specifier  : typedef_name
                            | enum_specifier
                            | struct_or_union_specifier
                            | type_specifier_no_typeid
        """
        p[0] = p[1]

    def p_type_qualifier(self, p):
        """ type_qualifier  : CONST
                            | RESTRICT
                            | VOLATILE
        """
        p[0] = p[1]

    def p_init_declarator_list(self, p):
        """ init_declarator_list    : init_declarator
                                    | init_declarator_list COMMA init_declarator
        """
        p[0] = p[1] + [p[3]] if len(p) == 4 else [p[1]]

    # Returns a {decl=<declarator> : init=<initializer>} dictionary
    # If there's no initializer, uses None
    #
    def p_init_declarator(self, p):
        """ init_declarator : declarator
                            | declarator EQUALS initializer
        """
        p[0] = dict(decl=p[1], init=(p[3] if len(p) > 2 else None))

    def p_id_init_declarator_list(self, p):
        """ id_init_declarator_list    : id_init_declarator
                                       | id_init_declarator_list COMMA init_declarator
        """
        p[0] = p[1] + [p[3]] if len(p) == 4 else [p[1]]

    def p_id_init_declarator(self, p):
        """ id_init_declarator : id_declarator
                               | id_declarator EQUALS initializer
        """
        p[0] = dict(decl=p[1], init=(p[3] if len(p) > 2 else None))

    # Require at least one type specifier in a specifier-qualifier-list
    #
    def p_specifier_qualifier_list_1(self, p):
        """ specifier_qualifier_list    : specifier_qualifier_list type_specifier_no_typeid
        """
        p[0] = self._add_declaration_specifier(p[1], p[2], 'type', append=True)

    def p_specifier_qualifier_list_2(self, p):
        """ specifier_qualifier_list    : specifier_qualifier_list type_qualifier
        """
        p[0] = self._add_declaration_specifier(p[1], p[2], 'qual', append=True)

    def p_specifier_qualifier_list_3(self, p):
        """ specifier_qualifier_list  : type_specifier
        """
        p[0] = self._add_declaration_specifier(None, p[1], 'type')

    def p_specifier_qualifier_list_4(self, p):
        """ specifier_qualifier_list  : type_qualifier_list type_specifier
        """
        spec = dict(qual=p[1], storage=[], type=[], function=[])
        p[0] = self._add_declaration_specifier(spec, p[2], 'type', append=True)

    # TYPEID is allowed here (and in other struct/enum related tag names), because
    # struct/enum tags reside in their own namespace and can be named the same as types
    #
    def p_struct_or_union_specifier_1(self, p):
        """ struct_or_union_specifier   : struct_or_union ID
                                        | struct_or_union TYPEID
        """
        klass = self._select_struct_union_class(p[1])
        # None means no list of members
        p[0] = klass(
            name=p[2],
            decls=None,
            coord=self._token_coord(p, 2))

    def p_struct_or_union_specifier_2(self, p):
        """ struct_or_union_specifier : struct_or_union brace_open struct_declaration_list brace_close
                                      | struct_or_union brace_open brace_close
        """
        klass = self._select_struct_union_class(p[1])
        if len(p) == 4:
            # Empty sequence means an empty list of members
            p[0] = klass(
                name=None,
                decls=[],
                coord=self._token_coord(p, 2))
        else:
            p[0] = klass(
                name=None,
                decls=p[3],
                coord=self._token_coord(p, 2))


    def p_struct_or_union_specifier_3(self, p):
        """ struct_or_union_specifier   : struct_or_union ID brace_open struct_declaration_list brace_close
                                        | struct_or_union ID brace_open brace_close
                                        | struct_or_union TYPEID brace_open struct_declaration_list brace_close
                                        | struct_or_union TYPEID brace_open brace_close
        """
        klass = self._select_struct_union_class(p[1])
        if len(p) == 5:
            # Empty sequence means an empty list of members
            p[0] = klass(
                name=p[2],
                decls=[],
                coord=self._token_coord(p, 2))
        else:
            p[0] = klass(
                name=p[2],
                decls=p[4],
                coord=self._token_coord(p, 2))

    def p_struct_or_union(self, p):
        """ struct_or_union : STRUCT
                            | UNION
        """
        p[0] = p[1]

    # Combine all declarations into a single list
    #
    def p_struct_declaration_list(self, p):
        """ struct_declaration_list     : struct_declaration
                                        | struct_declaration_list struct_declaration
        """
        if len(p) == 2:
            p[0] = p[1] or []
        else:
            p[0] = p[1] + (p[2] or [])

    def p_struct_declaration_1(self, p):
        """ struct_declaration : specifier_qualifier_list struct_declarator_list_opt SEMI
        """
        spec = p[1]
        assert 'typedef' not in spec['storage']

        if p[2] is not None:
            decls = self._build_declarations(
                spec=spec,
                decls=p[2])

        elif len(spec['type']) == 1:
            # Anonymous struct/union, gcc extension, C1x feature.
            # Although the standard only allows structs/unions here, I see no
            # reason to disallow other types since some compilers have typedefs
            # here, and pycparser isn't about rejecting all invalid code.
            #
            node = spec['type'][0]
            if isinstance(node, c_ast.Node):
                decl_type = node
            else:
                decl_type = c_ast.IdentifierType(node)

            decls = self._build_declarations(
                spec=spec,
                decls=[dict(decl=decl_type)])

        else:
            # Structure/union members can have the same names as typedefs.
            # The trouble is that the member's name gets grouped into
            # specifier_qualifier_list; _build_declarations compensates.
            #
            decls = self._build_declarations(
                spec=spec,
                decls=[dict(decl=None, init=None)])

        p[0] = decls

    def p_struct_declaration_2(self, p):
        """ struct_declaration : SEMI
        """
        p[0] = None

    def p_struct_declaration_3(self, p):
        """ struct_declaration : pppragma_directive
        """
        p[0] = [p[1]]

    def p_struct_declarator_list(self, p):
        """ struct_declarator_list  : struct_declarator
                                    | struct_declarator_list COMMA struct_declarator
        """
        p[0] = p[1] + [p[3]] if len(p) == 4 else [p[1]]

    # struct_declarator passes up a dict with the keys: decl (for
    # the underlying declarator) and bitsize (for the bitsize)
    #
    def p_struct_declarator_1(self, p):
        """ struct_declarator : declarator
        """
        p[0] = {'decl': p[1], 'bitsize': None}

    def p_struct_declarator_2(self, p):
        """ struct_declarator   : declarator COLON constant_expression
                                | COLON constant_expression
        """
        if len(p) > 3:
            p[0] = {'decl': p[1], 'bitsize': p[3]}
        else:
            p[0] = {'decl': c_ast.TypeDecl(None, None, None), 'bitsize': p[2]}

    def p_enum_specifier_1(self, p):
        """ enum_specifier  : ENUM ID
                            | ENUM TYPEID
        """
        p[0] = c_ast.Enum(p[2], None, self._token_coord(p, 1))

    def p_enum_specifier_2(self, p):
        """ enum_specifier  : ENUM brace_open enumerator_list brace_close
        """
        p[0] = c_ast.Enum(None, p[3], self._token_coord(p, 1))

    def p_enum_specifier_3(self, p):
        """ enum_specifier  : ENUM ID brace_open enumerator_list brace_close
                            | ENUM TYPEID brace_open enumerator_list brace_close
        """
        p[0] = c_ast.Enum(p[2], p[4], self._token_coord(p, 1))

    def p_enumerator_list(self, p):
        """ enumerator_list : enumerator
                            | enumerator_list COMMA
                            | enumerator_list COMMA enumerator
        """
        if len(p) == 2:
            p[0] = c_ast.EnumeratorList([p[1]], p[1].coord)
        elif len(p) == 3:
            p[0] = p[1]
        else:
            p[1].enumerators.append(p[3])
            p[0] = p[1]

    def p_enumerator(self, p):
        """ enumerator  : ID
                        | ID EQUALS constant_expression
        """
        if len(p) == 2:
            enumerator = c_ast.Enumerator(
                        p[1], None,
                        self._token_coord(p, 1))
        else:
            enumerator = c_ast.Enumerator(
                        p[1], p[3],
                        self._token_coord(p, 1))
        self._add_identifier(enumerator.name, enumerator.coord)

        p[0] = enumerator

    def p_declarator(self, p):
        """ declarator  : id_declarator
                        | typeid_declarator
        """
        p[0] = p[1]

    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
    def p_xxx_declarator_1(self, p):
        """ xxx_declarator  : direct_xxx_declarator
        """
        p[0] = p[1]

    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
    def p_xxx_declarator_2(self, p):
        """ xxx_declarator  : pointer direct_xxx_declarator
        """
        p[0] = self._type_modify_decl(p[2], p[1])

    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
    def p_direct_xxx_declarator_1(self, p):
        """ direct_xxx_declarator   : yyy
        """
        p[0] = c_ast.TypeDecl(
            declname=p[1],
            type=None,
            quals=None,
            coord=self._token_coord(p, 1))

    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'))
    def p_direct_xxx_declarator_2(self, p):
        """ direct_xxx_declarator   : LPAREN xxx_declarator RPAREN
        """
        p[0] = p[2]

    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
    def p_direct_xxx_declarator_3(self, p):
        """ direct_xxx_declarator   : direct_xxx_declarator LBRACKET type_qualifier_list_opt assignment_expression_opt RBRACKET
        """
        quals = (p[3] if len(p) > 5 else []) or []
        # Accept dimension qualifiers
        # Per C99 6.7.5.3 p7
        arr = c_ast.ArrayDecl(
            type=None,
            dim=p[4] if len(p) > 5 else p[3],
            dim_quals=quals,
            coord=p[1].coord)

        p[0] = self._type_modify_decl(decl=p[1], modifier=arr)

    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
    def p_direct_xxx_declarator_4(self, p):
        """ direct_xxx_declarator   : direct_xxx_declarator LBRACKET STATIC type_qualifier_list_opt assignment_expression RBRACKET
                                    | direct_xxx_declarator LBRACKET type_qualifier_list STATIC assignment_expression RBRACKET
        """
        # Using slice notation for PLY objects doesn't work in Python 3 for the
        # version of PLY embedded with pycparser; see PLY Google Code issue 30.
        # Work around that here by listing the two elements separately.
        listed_quals = [item if isinstance(item, list) else [item]
            for item in [p[3],p[4]]]
        dim_quals = [qual for sublist in listed_quals for qual in sublist
            if qual is not None]
        arr = c_ast.ArrayDecl(
            type=None,
            dim=p[5],
            dim_quals=dim_quals,
            coord=p[1].coord)

        p[0] = self._type_modify_decl(decl=p[1], modifier=arr)

    # Special for VLAs
    #
    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
    def p_direct_xxx_declarator_5(self, p):
        """ direct_xxx_declarator   : direct_xxx_declarator LBRACKET type_qualifier_list_opt TIMES RBRACKET
        """
        arr = c_ast.ArrayDecl(
            type=None,
            dim=c_ast.ID(p[4], self._token_coord(p, 4)),
            dim_quals=p[3] if p[3] != None else [],
            coord=p[1].coord)

        p[0] = self._type_modify_decl(decl=p[1], modifier=arr)

    @parameterized(('id', 'ID'), ('typeid', 'TYPEID'), ('typeid_noparen', 'TYPEID'))
    def p_direct_xxx_declarator_6(self, p):
        """ direct_xxx_declarator   : direct_xxx_declarator LPAREN parameter_type_list RPAREN
                                    | direct_xxx_declarator LPAREN identifier_list_opt RPAREN
        """
        func = c_ast.FuncDecl(
            args=p[3],
            type=None,
            coord=p[1].coord)

        # To see why _get_yacc_lookahead_token is needed, consider:
        #   typedef char TT;
        #   void foo(int TT) { TT = 10; }
        # Outside the function, TT is a typedef, but inside (starting and
        # ending with the braces) it's a parameter.  The trouble begins with
        # yacc's lookahead token.  We don't know if we're declaring or
        # defining a function until we see LBRACE, but if we wait for yacc to
        # trigger a rule on that token, then TT will have already been read
        # and incorrectly interpreted as TYPEID.  We need to add the
        # parameters to the scope the moment the lexer sees LBRACE.
        #
        if self._get_yacc_lookahead_token().type == "LBRACE":
            if func.args is not None:
                for param in func.args.params:
                    if isinstance(param, c_ast.EllipsisParam): break
                    self._add_identifier(param.name, param.coord)

        p[0] = self._type_modify_decl(decl=p[1], modifier=func)

    def p_pointer(self, p):
        """ pointer : TIMES type_qualifier_list_opt
                    | TIMES type_qualifier_list_opt pointer
        """
        coord = self._token_coord(p, 1)
        # Pointer decls nest from inside out. This is important when different
        # levels have different qualifiers. For example:
        #
        #  char * const * p;
        #
        # Means "pointer to const pointer to char"
        #
        # While:
        #
        #  char ** const p;
        #
        # Means "const pointer to pointer to char"
        #
        # So when we construct PtrDecl nestings, the leftmost pointer goes in
        # as the most nested type.
        nested_type = c_ast.PtrDecl(quals=p[2] or [], type=None, coord=coord)
        if len(p) > 3:
            tail_type = p[3]
            while tail_type.type is not None:
                tail_type = tail_type.type
            tail_type.type = nested_type
            p[0] = p[3]
        else:
            p[0] = nested_type

    def p_type_qualifier_list(self, p):
        """ type_qualifier_list : type_qualifier
                                | type_qualifier_list type_qualifier
        """
        p[0] = [p[1]] if len(p) == 2 else p[1] + [p[2]]

    def p_parameter_type_list(self, p):
        """ parameter_type_list : parameter_list
                                | parameter_list COMMA ELLIPSIS
        """
        if len(p) > 2:
            p[1].params.append(c_ast.EllipsisParam(self._token_coord(p, 3)))

        p[0] = p[1]

    def p_parameter_list(self, p):
        """ parameter_list  : parameter_declaration
                            | parameter_list COMMA parameter_declaration
        """
        if len(p) == 2: # single parameter
            p[0] = c_ast.ParamList([p[1]], p[1].coord)
        else:
            p[1].params.append(p[3])
            p[0] = p[1]

    # From ISO/IEC 9899:TC2, 6.7.5.3.11:
    # "If, in a parameter declaration, an identifier can be treated either
    #  as a typedef name or as a parameter name, it shall be taken as a
    #  typedef name."
    #
    # Inside a parameter declaration, once we've reduced declaration specifiers,
    # if we shift in an LPAREN and see a TYPEID, it could be either an abstract
    # declarator or a declarator nested inside parens. This rule tells us to
    # always treat it as an abstract declarator. Therefore, we only accept
    # `id_declarator`s and `typeid_noparen_declarator`s.
    def p_parameter_declaration_1(self, p):
        """ parameter_declaration   : declaration_specifiers id_declarator
                                    | declaration_specifiers typeid_noparen_declarator
        """
        spec = p[1]
        if not spec['type']:
            spec['type'] = [c_ast.IdentifierType(['int'],
                coord=self._token_coord(p, 1))]
        p[0] = self._build_declarations(
            spec=spec,
            decls=[dict(decl=p[2])])[0]

    def p_parameter_declaration_2(self, p):
        """ parameter_declaration   : declaration_specifiers abstract_declarator_opt
        """
        spec = p[1]
        if not spec['type']:
            spec['type'] = [c_ast.IdentifierType(['int'],
                coord=self._token_coord(p, 1))]

        # Parameters can have the same names as typedefs.  The trouble is that
        # the parameter's name gets grouped into declaration_specifiers, making
        # it look like an old-style declaration; compensate.
        #
        if len(spec['type']) > 1 and len(spec['type'][-1].names) == 1 and \
                self._is_type_in_scope(spec['type'][-1].names[0]):
            decl = self._build_declarations(
                    spec=spec,
                    decls=[dict(decl=p[2], init=None)])[0]

        # This truly is an old-style parameter declaration
        #
        else:
            decl = c_ast.Typename(
                name='',
                quals=spec['qual'],
                type=p[2] or c_ast.TypeDecl(None, None, None),
                coord=self._token_coord(p, 2))
            typename = spec['type']
            decl = self._fix_decl_name_type(decl, typename)

        p[0] = decl

    def p_identifier_list(self, p):
        """ identifier_list : identifier
                            | identifier_list COMMA identifier
        """
        if len(p) == 2: # single parameter
            p[0] = c_ast.ParamList([p[1]], p[1].coord)
        else:
            p[1].params.append(p[3])
            p[0] = p[1]

    def p_initializer_1(self, p):
        """ initializer : assignment_expression
        """
        p[0] = p[1]

    def p_initializer_2(self, p):
        """ initializer : brace_open initializer_list_opt brace_close
                        | brace_open initializer_list COMMA brace_close
        """
        if p[2] is None:
            p[0] = c_ast.InitList([], self._token_coord(p, 1))
        else:
            p[0] = p[2]

    def p_initializer_list(self, p):
        """ initializer_list    : designation_opt initializer
                                | initializer_list COMMA designation_opt initializer
        """
        if len(p) == 3: # single initializer
            init = p[2] if p[1] is None else c_ast.NamedInitializer(p[1], p[2])
            p[0] = c_ast.InitList([init], p[2].coord)
        else:
            init = p[4] if p[3] is None else c_ast.NamedInitializer(p[3], p[4])
            p[1].exprs.append(init)
            p[0] = p[1]

    def p_designation(self, p):
        """ designation : designator_list EQUALS
        """
        p[0] = p[1]

    # Designators are represented as a list of nodes, in the order in which
    # they're written in the code.
    #
    def p_designator_list(self, p):
        """ designator_list : designator
                            | designator_list designator
        """
        p[0] = [p[1]] if len(p) == 2 else p[1] + [p[2]]

    def p_designator(self, p):
        """ designator  : LBRACKET constant_expression RBRACKET
                        | PERIOD identifier
        """
        p[0] = p[2]

    def p_type_name(self, p):
        """ type_name   : specifier_qualifier_list abstract_declarator_opt
        """
        typename = c_ast.Typename(
            name='',
            quals=p[1]['qual'],
            type=p[2] or c_ast.TypeDecl(None, None, None),
            coord=self._token_coord(p, 2))

        p[0] = self._fix_decl_name_type(typename, p[1]['type'])

    def p_abstract_declarator_1(self, p):
        """ abstract_declarator     : pointer
        """
        dummytype = c_ast.TypeDecl(None, None, None)
        p[0] = self._type_modify_decl(
            decl=dummytype,
            modifier=p[1])

    def p_abstract_declarator_2(self, p):
        """ abstract_declarator     : pointer direct_abstract_declarator
        """
        p[0] = self._type_modify_decl(p[2], p[1])

    def p_abstract_declarator_3(self, p):
        """ abstract_declarator     : direct_abstract_declarator
        """
        p[0] = p[1]

    # Creating and using direct_abstract_declarator_opt here
    # instead of listing both direct_abstract_declarator and the
    # lack of it in the beginning of _1 and _2 caused two
    # shift/reduce errors.
    #
    def p_direct_abstract_declarator_1(self, p):
        """ direct_abstract_declarator  : LPAREN abstract_declarator RPAREN """
        p[0] = p[2]

    def p_direct_abstract_declarator_2(self, p):
        """ direct_abstract_declarator  : direct_abstract_declarator LBRACKET assignment_expression_opt RBRACKET
        """
        arr = c_ast.ArrayDecl(
            type=None,
            dim=p[3],
            dim_quals=[],
            coord=p[1].coord)

        p[0] = self._type_modify_decl(decl=p[1], modifier=arr)

    def p_direct_abstract_declarator_3(self, p):
        """ direct_abstract_declarator  : LBRACKET assignment_expression_opt RBRACKET
        """
        p[0] = c_ast.ArrayDecl(
            type=c_ast.TypeDecl(None, None, None),
            dim=p[2],
            dim_quals=[],
            coord=self._token_coord(p, 1))

    def p_direct_abstract_declarator_4(self, p):
        """ direct_abstract_declarator  : direct_abstract_declarator LBRACKET TIMES RBRACKET
        """
        arr = c_ast.ArrayDecl(
            type=None,
            dim=c_ast.ID(p[3], self._token_coord(p, 3)),
            dim_quals=[],
            coord=p[1].coord)

        p[0] = self._type_modify_decl(decl=p[1], modifier=arr)

    def p_direct_abstract_declarator_5(self, p):
        """ direct_abstract_declarator  : LBRACKET TIMES RBRACKET
        """
        p[0] = c_ast.ArrayDecl(
            type=c_ast.TypeDecl(None, None, None),
            dim=c_ast.ID(p[3], self._token_coord(p, 3)),
            dim_quals=[],
            coord=self._token_coord(p, 1))

    def p_direct_abstract_declarator_6(self, p):
        """ direct_abstract_declarator  : direct_abstract_declarator LPAREN parameter_type_list_opt RPAREN
        """
        func = c_ast.FuncDecl(
            args=p[3],
            type=None,
            coord=p[1].coord)

        p[0] = self._type_modify_decl(decl=p[1], modifier=func)

    def p_direct_abstract_declarator_7(self, p):
        """ direct_abstract_declarator  : LPAREN parameter_type_list_opt RPAREN
        """
        p[0] = c_ast.FuncDecl(
            args=p[2],
            type=c_ast.TypeDecl(None, None, None),
            coord=self._token_coord(p, 1))

    # declaration is a list, statement isn't. To make it consistent, block_item
    # will always be a list
    #
    def p_block_item(self, p):
        """ block_item  : declaration
                        | statement
        """
        p[0] = p[1] if isinstance(p[1], list) else [p[1]]

    # Since we made block_item a list, this just combines lists
    #
    def p_block_item_list(self, p):
        """ block_item_list : block_item
                            | block_item_list block_item
        """
        # Empty block items (plain ';') produce [None], so ignore them
        p[0] = p[1] if (len(p) == 2 or p[2] == [None]) else p[1] + p[2]

    def p_compound_statement_1(self, p):
        """ compound_statement : brace_open block_item_list_opt brace_close """
        p[0] = c_ast.Compound(
            block_items=p[2],
            coord=self._token_coord(p, 1))

    def p_labeled_statement_1(self, p):
        """ labeled_statement : ID COLON pragmacomp_or_statement """
        p[0] = c_ast.Label(p[1], p[3], self._token_coord(p, 1))

    def p_labeled_statement_2(self, p):
        """ labeled_statement : CASE constant_expression COLON pragmacomp_or_statement """
        p[0] = c_ast.Case(p[2], [p[4]], self._token_coord(p, 1))

    def p_labeled_statement_3(self, p):
        """ labeled_statement : DEFAULT COLON pragmacomp_or_statement """
        p[0] = c_ast.Default([p[3]], self._token_coord(p, 1))

    def p_selection_statement_1(self, p):
        """ selection_statement : IF LPAREN expression RPAREN pragmacomp_or_statement """
        p[0] = c_ast.If(p[3], p[5], None, self._token_coord(p, 1))

    def p_selection_statement_2(self, p):
        """ selection_statement : IF LPAREN expression RPAREN statement ELSE pragmacomp_or_statement """
        p[0] = c_ast.If(p[3], p[5], p[7], self._token_coord(p, 1))

    def p_selection_statement_3(self, p):
        """ selection_statement : SWITCH LPAREN expression RPAREN pragmacomp_or_statement """
        p[0] = fix_switch_cases(
                c_ast.Switch(p[3], p[5], self._token_coord(p, 1)))

    def p_iteration_statement_1(self, p):
        """ iteration_statement : WHILE LPAREN expression RPAREN pragmacomp_or_statement """
        p[0] = c_ast.While(p[3], p[5], self._token_coord(p, 1))

    def p_iteration_statement_2(self, p):
        """ iteration_statement : DO pragmacomp_or_statement WHILE LPAREN expression RPAREN SEMI """
        p[0] = c_ast.DoWhile(p[5], p[2], self._token_coord(p, 1))

    def p_iteration_statement_3(self, p):
        """ iteration_statement : FOR LPAREN expression_opt SEMI expression_opt SEMI expression_opt RPAREN pragmacomp_or_statement """
        p[0] = c_ast.For(p[3], p[5], p[7], p[9], self._token_coord(p, 1))

    def p_iteration_statement_4(self, p):
        """ iteration_statement : FOR LPAREN declaration expression_opt SEMI expression_opt RPAREN pragmacomp_or_statement """
        p[0] = c_ast.For(c_ast.DeclList(p[3], self._token_coord(p, 1)),
                         p[4], p[6], p[8], self._token_coord(p, 1))

    def p_jump_statement_1(self, p):
        """ jump_statement  : GOTO ID SEMI """
        p[0] = c_ast.Goto(p[2], self._token_coord(p, 1))

    def p_jump_statement_2(self, p):
        """ jump_statement  : BREAK SEMI """
        p[0] = c_ast.Break(self._token_coord(p, 1))

    def p_jump_statement_3(self, p):
        """ jump_statement  : CONTINUE SEMI """
        p[0] = c_ast.Continue(self._token_coord(p, 1))

    def p_jump_statement_4(self, p):
        """ jump_statement  : RETURN expression SEMI
                            | RETURN SEMI
        """
        p[0] = c_ast.Return(p[2] if len(p) == 4 else None, self._token_coord(p, 1))

    def p_expression_statement(self, p):
        """ expression_statement : expression_opt SEMI """
        if p[1] is None:
            p[0] = c_ast.EmptyStatement(self._token_coord(p, 2))
        else:
            p[0] = p[1]

    def p_expression(self, p):
        """ expression  : assignment_expression
                        | expression COMMA assignment_expression
        """
        if len(p) == 2:
            p[0] = p[1]
        else:
            if not isinstance(p[1], c_ast.ExprList):
                p[1] = c_ast.ExprList([p[1]], p[1].coord)

            p[1].exprs.append(p[3])
            p[0] = p[1]

    def p_typedef_name(self, p):
        """ typedef_name : TYPEID """
        p[0] = c_ast.IdentifierType([p[1]], coord=self._token_coord(p, 1))

    def p_assignment_expression(self, p):
        """ assignment_expression   : conditional_expression
                                    | unary_expression assignment_operator assignment_expression
        """
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = c_ast.Assignment(p[2], p[1], p[3], p[1].coord)

    # K&R2 defines these as many separate rules, to encode
    # precedence and associativity. Why work hard ? I'll just use
    # the built in precedence/associativity specification feature
    # of PLY. (see precedence declaration above)
    #
    def p_assignment_operator(self, p):
        """ assignment_operator : EQUALS
                                | XOREQUAL
                                | TIMESEQUAL
                                | DIVEQUAL
                                | MODEQUAL
                                | PLUSEQUAL
                                | MINUSEQUAL
                                | LSHIFTEQUAL
                                | RSHIFTEQUAL
                                | ANDEQUAL
                                | OREQUAL
        """
        p[0] = p[1]

    def p_constant_expression(self, p):
        """ constant_expression : conditional_expression """
        p[0] = p[1]

    def p_conditional_expression(self, p):
        """ conditional_expression  : binary_expression
                                    | binary_expression CONDOP expression COLON conditional_expression
        """
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = c_ast.TernaryOp(p[1], p[3], p[5], p[1].coord)

    def p_binary_expression(self, p):
        """ binary_expression   : cast_expression
                                | binary_expression TIMES binary_expression
                                | binary_expression DIVIDE binary_expression
                                | binary_expression MOD binary_expression
                                | binary_expression PLUS binary_expression
                                | binary_expression MINUS binary_expression
                                | binary_expression RSHIFT binary_expression
                                | binary_expression LSHIFT binary_expression
                                | binary_expression LT binary_expression
                                | binary_expression LE binary_expression
                                | binary_expression GE binary_expression
                                | binary_expression GT binary_expression
                                | binary_expression EQ binary_expression
                                | binary_expression NE binary_expression
                                | binary_expression AND binary_expression
                                | binary_expression OR binary_expression
                                | binary_expression XOR binary_expression
                                | binary_expression LAND binary_expression
                                | binary_expression LOR binary_expression
        """
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = c_ast.BinaryOp(p[2], p[1], p[3], p[1].coord)

    def p_cast_expression_1(self, p):
        """ cast_expression : unary_expression """
        p[0] = p[1]

    def p_cast_expression_2(self, p):
        """ cast_expression : LPAREN type_name RPAREN cast_expression """
        p[0] = c_ast.Cast(p[2], p[4], self._token_coord(p, 1))

    def p_unary_expression_1(self, p):
        """ unary_expression    : postfix_expression """
        p[0] = p[1]

    def p_unary_expression_2(self, p):
        """ unary_expression    : PLUSPLUS unary_expression
                                | MINUSMINUS unary_expression
                                | unary_operator cast_expression
        """
        p[0] = c_ast.UnaryOp(p[1], p[2], p[2].coord)

    def p_unary_expression_3(self, p):
        """ unary_expression    : SIZEOF unary_expression
                                | SIZEOF LPAREN type_name RPAREN
        """
        p[0] = c_ast.UnaryOp(
            p[1],
            p[2] if len(p) == 3 else p[3],
            self._token_coord(p, 1))

    def p_unary_operator(self, p):
        """ unary_operator  : AND
                            | TIMES
                            | PLUS
                            | MINUS
                            | NOT
                            | LNOT
        """
        p[0] = p[1]

    def p_postfix_expression_1(self, p):
        """ postfix_expression  : primary_expression """
        p[0] = p[1]

    def p_postfix_expression_2(self, p):
        """ postfix_expression  : postfix_expression LBRACKET expression RBRACKET """
        p[0] = c_ast.ArrayRef(p[1], p[3], p[1].coord)

    def p_postfix_expression_3(self, p):
        """ postfix_expression  : postfix_expression LPAREN argument_expression_list RPAREN
                                | postfix_expression LPAREN RPAREN
        """
        p[0] = c_ast.FuncCall(p[1], p[3] if len(p) == 5 else None, p[1].coord)

    def p_postfix_expression_4(self, p):
        """ postfix_expression  : postfix_expression PERIOD ID
                                | postfix_expression PERIOD TYPEID
                                | postfix_expression ARROW ID
                                | postfix_expression ARROW TYPEID
        """
        field = c_ast.ID(p[3], self._token_coord(p, 3))
        p[0] = c_ast.StructRef(p[1], p[2], field, p[1].coord)

    def p_postfix_expression_5(self, p):
        """ postfix_expression  : postfix_expression PLUSPLUS
                                | postfix_expression MINUSMINUS
        """
        p[0] = c_ast.UnaryOp('p' + p[2], p[1], p[1].coord)

    def p_postfix_expression_6(self, p):
        """ postfix_expression  : LPAREN type_name RPAREN brace_open initializer_list brace_close
                                | LPAREN type_name RPAREN brace_open initializer_list COMMA brace_close
        """
        p[0] = c_ast.CompoundLiteral(p[2], p[5])

    def p_primary_expression_1(self, p):
        """ primary_expression  : identifier """
        p[0] = p[1]

    def p_primary_expression_2(self, p):
        """ primary_expression  : constant """
        p[0] = p[1]

    def p_primary_expression_3(self, p):
        """ primary_expression  : unified_string_literal
                                | unified_wstring_literal
        """
        p[0] = p[1]

    def p_primary_expression_4(self, p):
        """ primary_expression  : LPAREN expression RPAREN """
        p[0] = p[2]

    def p_primary_expression_5(self, p):
        """ primary_expression  : OFFSETOF LPAREN type_name COMMA offsetof_member_designator RPAREN
        """
        coord = self._token_coord(p, 1)
        p[0] = c_ast.FuncCall(c_ast.ID(p[1], coord),
                              c_ast.ExprList([p[3], p[5]], coord),
                              coord)

    def p_offsetof_member_designator(self, p):
        """ offsetof_member_designator : identifier
                                         | offsetof_member_designator PERIOD identifier
                                         | offsetof_member_designator LBRACKET expression RBRACKET
        """
        if len(p) == 2:
            p[0] = p[1]
        elif len(p) == 4:
            field = c_ast.ID(p[3], self._token_coord(p, 3))
            p[0] = c_ast.StructRef(p[1], p[2], field, p[1].coord)
        elif len(p) == 5:
            p[0] = c_ast.ArrayRef(p[1], p[3], p[1].coord)
        else:
            raise NotImplementedError("Unexpected parsing state. len(p): %u" % len(p))

    def p_argument_expression_list(self, p):
        """ argument_expression_list    : assignment_expression
                                        | argument_expression_list COMMA assignment_expression
        """
        if len(p) == 2: # single expr
            p[0] = c_ast.ExprList([p[1]], p[1].coord)
        else:
            p[1].exprs.append(p[3])
            p[0] = p[1]

    def p_identifier(self, p):
        """ identifier  : ID """
        p[0] = c_ast.ID(p[1], self._token_coord(p, 1))

    def p_constant_1(self, p):
        """ constant    : INT_CONST_DEC
                        | INT_CONST_OCT
                        | INT_CONST_HEX
                        | INT_CONST_BIN
        """
        p[0] = c_ast.Constant(
            'int', p[1], self._token_coord(p, 1))

    def p_constant_2(self, p):
        """ constant    : FLOAT_CONST
                        | HEX_FLOAT_CONST
        """
        if 'x' in p[1].lower():
            t = 'float'
        else:
            if p[1][-1] in ('f', 'F'):
                t = 'float'
            elif p[1][-1] in ('l', 'L'):
                t = 'long double'
            else:
                t = 'double'

        p[0] = c_ast.Constant(
            t, p[1], self._token_coord(p, 1))

    def p_constant_3(self, p):
        """ constant    : CHAR_CONST
                        | WCHAR_CONST
        """
        p[0] = c_ast.Constant(
            'char', p[1], self._token_coord(p, 1))

    # The "unified" string and wstring literal rules are for supporting
    # concatenation of adjacent string literals.
    # I.e. "hello " "world" is seen by the C compiler as a single string literal
    # with the value "hello world"
    #
    def p_unified_string_literal(self, p):
        """ unified_string_literal  : STRING_LITERAL
                                    | unified_string_literal STRING_LITERAL
        """
        if len(p) == 2: # single literal
            p[0] = c_ast.Constant(
                'string', p[1], self._token_coord(p, 1))
        else:
            p[1].value = p[1].value[:-1] + p[2][1:]
            p[0] = p[1]

    def p_unified_wstring_literal(self, p):
        """ unified_wstring_literal : WSTRING_LITERAL
                                    | unified_wstring_literal WSTRING_LITERAL
        """
        if len(p) == 2: # single literal
            p[0] = c_ast.Constant(
                'string', p[1], self._token_coord(p, 1))
        else:
            p[1].value = p[1].value.rstrip()[:-1] + p[2][2:]
            p[0] = p[1]

    def p_brace_open(self, p):
        """ brace_open  :   LBRACE
        """
        p[0] = p[1]
        p.set_lineno(0, p.lineno(1))

    def p_brace_close(self, p):
        """ brace_close :   RBRACE
        """
        p[0] = p[1]
        p.set_lineno(0, p.lineno(1))

    def p_empty(self, p):
        'empty : '
        p[0] = None

    def p_error(self, p):
        # If error recovery is added here in the future, make sure
        # _get_yacc_lookahead_token still works!
        #
        if p:
            self._parse_error(
                'before: %s' % p.value,
                self._coord(lineno=p.lineno,
                            column=self.clex.find_tok_column(p)))
        else:
            self._parse_error('At end of input', self.clex.filename)



from report.command import AlignCommand, BoldCommand

from .._CommandProcessor import CommandProcessor


class BoldCommandProcessor(CommandProcessor):
    def get_bytes(self, command):
        if not isinstance(command, BoldCommand):
            return ""
        
        ret = b"\x45"
        if command.activate:
            ret += "\x00"
        else:
            ret += "\x01"

        return ret



# -*- coding: utf-8 -*-

import json

from sysactions import action, get_model, get_posot, show_messagebox, close_asynch_dialog

from utilfunctions import compare_objects

from ..customer import Customer
from ..customer.customer_model import AddCustomerRequest

dialog_id = None


def insert_or_update_customer(pos_id, customer_json):
    # type: (int, str) -> None

    model = get_model(pos_id)
    pos_ot = get_posot(model)
    _update_order_custom_properties(customer_json, model, pos_id, pos_ot)

    customer_request = _create_customer_request(customer_json)

    if customer_request.phone:
        customer = Customer().get_customer(customer_request.phone)
        if not customer:
            message = "$CUSTOMER_WILL_BE_INSERTED"
            confirmation = show_messagebox(pos_id, message, timeout=720000, buttons="$OK|$CANCEL")
            if confirmation == 1:
                return

            Customer().add_customer(customer_request)
            return

        customer_updated = compare_objects(customer_request, _create_customer_request_to_compare(json.loads(customer)))
        if not customer_updated:
            message = "$CUSTOMER_WILL_BE_UPDATED"
            confirmation = show_messagebox(pos_id, message, timeout=720000, buttons="$YES|$NO")
            if confirmation == 1:
                return

            Customer().update_customer(customer_request)


def _create_customer_request(address_data_load):
    add_customer_request = AddCustomerRequest(
            address_data_load["CUSTOMER_PHONE"],
            address_data_load["CUSTOMER_DOC"],
            address_data_load["CUSTOMER_NAME"],
            address_data_load["POSTAL_CODE"],
            address_data_load["CITY"],
            address_data_load["STREET_NAME"],
            address_data_load["NEIGHBORHOOD"],
            address_data_load["STREET_NUMBER"],
            address_data_load["COMPLEMENT"],
            address_data_load["ADDRESS_REFERENCE"],
    )

    return add_customer_request


def _create_customer_request_to_compare(customer):
    add_customer_request = AddCustomerRequest(
            customer["phone"],
            customer["document"],
            customer["name"],
            customer["address"]["zip_code"],
            customer["address"]["city"],
            customer["address"]["street"],
            customer["address"]["neighborhood"],
            customer["address"]["number"],
            customer["address"]["complement"],
            customer["address"]["reference_point"]
    )
    return add_customer_request


@action
def zip_code_search(pos_id):
    global dialog_id
    dialog_id = show_messagebox(pos_id, "$SEARCHING_ZIP_CODE", buttons="", asynch=True, timeout=50000)


@action
def zip_code_search_finished(pos_id, error="false"):
    close_asynch_dialog(pos_id, dialog_id)
    if error == "true":
        show_messagebox(pos_id, "$ERROR_SEARCHING_ZIP_CODE", buttons="OK")


def _update_order_custom_properties(address_data_load, model, pos_id, pos_ot):
    for key, value in address_data_load.iteritems():
        if value == '':
            if model.find('.//OrderProperty[@key="{}"]'.format(key)) is not None:
                pos_ot.setOrderCustomProperty(key, 'clear')
        else:
            pos_ot.setOrderCustomProperty(key, value.encode('utf-8'))

    _update_order_properties(pos_id, pos_ot, model)


def _update_order_properties(pos_id, pos_ot, model):
    sale_type = model.find(".//Order").get("saleType")
    pos_ot.updateOrderProperties(pos_id, saletype=sale_type)



from ._OrderPicture import OrderPicture
from ._TablePicture import TablePicture


class Order(object):
    def __init__(
        self,
        store_id,   # type: str
    ):
        # type: (...) -> None

        self.order_picture = OrderPicture(store_id)
        self.table_picture = TablePicture(store_id)




import pyscripts
import reports
import sysactions

pyscripts.mbcontext = reports.mbcontext
import tablemgrapi

from domain import TableService

tip_rate = 0


class MsgBusTableService(TableService):

    def get_table_picture(self, pos_id, table_id):
        pyscripts.mbcontext = reports.mbcontext
        sysactions.mbcontext = reports.mbcontext
        tablemgrapi.mbcontext = reports.mbcontext

        model = sysactions.get_model(pos_id)
        posts = tablemgrapi.get_pos_ts(model)
        return posts.getTablePicture(table_id)

    def get_tip_rate(self):
        global tip_rate
        if not tip_rate:
            tip_rate = sysactions.get_storewide_config("Store.TipRate", defval="10")
        return float(tip_rate)



from typing import Dict, Any
from pos_model import Order
from _NfceXmlPartBuilder import NfceXmlPartBuilder
from _ContextKeys import ContextKeys


class TotalBuilder(NfceXmlPartBuilder):
    def __init__(self, versao_ws):
        # type: (int) -> None
        self.versao_ws = versao_ws

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode
        v_prod = context[ContextKeys.total_prod]
        discount_value = float(order.discount_amount or 0)
        v_desc = discount_value if discount_value > 0 else 0
        v_outro = abs(discount_value) if discount_value < 0 else 0
        
        context[ContextKeys.v_nf] = v_nf = v_prod - discount_value
        
        v_tot_trib = context[ContextKeys.total_icms] + \
                     context[ContextKeys.total_pis] + \
                     context[ContextKeys.total_cofins]

        xml = "<total>" \
              "<ICMSTot>" \
              "<vBC>{v_bc:.02f}</vBC>" \
              "<vICMS>{v_icms:.02f}</vICMS>" \
              "<vICMSDeson>{v_icms_deson:.02f}</vICMSDeson>".format(v_bc=context[ContextKeys.total_icms_vbc],
                                                                    v_icms=context[ContextKeys.total_icms],
                                                                    v_icms_deson=context.get(ContextKeys.deson, 0))
        if self.versao_ws not in (1, 3):
            xml += "<vFCP>{v_fcp:.02f}</vFCP>".format(v_fcp=context.get(ContextKeys.total_fcp, 0))

        xml += "<vBCST>{v_bcst:.02f}</vBCST>" \
               "<vST>{v_st:.02f}</vST>".format(v_bcst=0, v_st=0)

        if self.versao_ws not in (1, 3):
            xml += "<vFCPST>{v_fcpst:.02f}</vFCPST>" \
                   "<vFCPSTRet>{v_fcpst_ret:.02f}</vFCPSTRet>".format(v_fcpst=context.get(ContextKeys.total_fcp_st, 0),
                                                                      v_fcpst_ret=context.get(ContextKeys.total_fcp_ant, 0))
        xml += "<vProd>{v_prod:.02f}</vProd>" \
               "<vFrete>{v_frete:.02f}</vFrete>" \
               "<vSeg>{v_seg:.02f}</vSeg>" \
               "<vDesc>{v_desc:.02f}</vDesc>" \
               "<vII>{v_ii:.02f}</vII>" \
               "<vIPI>{v_ipi:.02f}</vIPI>".format(v_prod=v_prod,
                                                  v_frete=0,
                                                  v_seg=0,
                                                  v_desc=v_desc,
                                                  v_ii=0,
                                                  v_ipi=0)
        if self.versao_ws not in (1, 3):
            xml += "<vIPIDevol>{v_ipi_devolv:.02f}</vIPIDevol>".format(v_ipi_devolv=0)
            
        xml += "<vPIS>{v_pis:.02f}</vPIS>" \
               "<vCOFINS>{v_cofins:.02f}</vCOFINS>" \
               "<vOutro>{v_outro:.02f}</vOutro>" \
               "<vNF>{v_nf:.02f}</vNF>" \
               "<vTotTrib>{v_tot_trib:.02f}</vTotTrib>" \
               "</ICMSTot>" \
               "</total>".format(v_pis=context[ContextKeys.total_pis],
                                 v_cofins=context[ContextKeys.total_cofins],
                                 v_outro=v_outro,
                                 v_nf=v_nf,
                                 v_tot_trib=v_tot_trib)

        return xml



# encoding: utf-8

from typing import Dict


class OrderBenefitInfo(object):
    def __init__(self, benefit,
                 discount_item_values,
                 order_state,
                 benefit_key,
                 promotion_tender,
                 burnt_items,
                 benefit_total_value):
        # type: (Dict, Dict, str, str, Dict, Dict, float) -> None

        self.benefit = benefit
        self.order_state = order_state
        self.discount_item_values = discount_item_values
        self.benefit_key = benefit_key
        self.burnt_items = burnt_items
        self.promotion_tender = promotion_tender
        self.benefit_total_value = benefit_total_value
        self.benefit_name = self.benefit.get("name", "").encode("utf8")
        self.benefit_description = self.benefit.get("userDescription", "").encode("utf8")

    @staticmethod
    def from_json(json):
        # type: (Dict) -> OrderBenefitInfo

        benefit = json.get("benefit", {})
        order_state = json.get("order_state")
        item_values = json.get("discount_item_values", {})
        benefit_key = json.get("benefit_key", None)
        burnt_items = json.get("burnt_items", {})
        promotion_tender = json.get("promotion_tender", [])
        benefit_total_value = json.get("benefit_total_value", 0.0)

        return OrderBenefitInfo(benefit,
                                item_values,
                                order_state,
                                benefit_key,
                                promotion_tender,
                                burnt_items,
                                benefit_total_value)



import base64
import json
from xml.etree import cElementTree as eTree

import sysactions
from modelcommon.printer import SeparatorTag
from msgbus import (
    FM_PARAM,
    TK_PRN_PRINT
)
from sysactions import (
    get_storewide_config,
    get_model,
    get_current_order,
    show_confirmation,
    get_podtype
)
from typing import (
    Dict,
    Optional
)
from utilfunctions import get_device_service


class PrintReceipt(object):
    def __init__(self, mb_context, logger):
        self.mb_context = mb_context
        self.logger = logger

        self.print_merchant_receipt = self._get_store_config("Store.PrintMerchantReceipt", "False")
        self.print_customer_receipt = self._get_store_config("Store.PrintCustomerReceipt", "False")
        self.ask_print_customer_receipt = self._get_store_config("Store.AskPrintCustomerReceipt", "True")

    @staticmethod
    def _get_store_config(
        config,         # type: str
        default_value,  # type: str
    ):
        # type: (...) -> bool

        return get_storewide_config(config, defval=default_value).lower() == "true"

    def print_tef_receipts(
        self,
        pos_id,                 # type: str
        fiscalization_type="",  # type: Optional[str]
        cancel_receipt=False    # type: Optional[bool]
    ):
        # type: (...) -> None

        try:
            model = get_model(pos_id)
            order_id = get_current_order(model).get("orderId")
            order_taker = sysactions.get_posot(model)
            order_pict = order_taker.orderPicture(orderid=order_id)
            order = eTree.XML(order_pict).find("Order")
            tenders = order.findall("TenderHistory/Tender")
            print_customer_receipt_response = None

            sector_name = None
            sector = order.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format("SECTOR"))
            if sector is not None:
                sector_name = sector.get("value", None)

            for tender in tenders:
                if int(tender.get("eletronicType")) not in (0, 9):
                    tender_detail = json.loads(tender.get("tenderDetail"))
                    print_customer_receipt_response = self.print_tef_tender_receipt(
                        pos_id=pos_id,
                        tender_detail=tender_detail,
                        fiscalization_type=fiscalization_type,
                        cancel_receipt=cancel_receipt,
                        print_customer_receipt=print_customer_receipt_response,
                        sector=sector_name,
                    )
        except Exception:  # noqa
            self.logger.exception("Error printing TEF receipts")

    def print_tef_tender_receipt(
        self,
        pos_id,                         # type: str
        tender_detail,                  # type: Dict
        fiscalization_type="",          # type: Optional[str]
        cancel_receipt=False,           # type: Optional[bool]
        print_customer_receipt=None,    # type: Optional[bool]
        sector=None,                    # type: str
    ):
        # type: (...) -> Optional[bool]

        container = tender_detail.get("ReceiptCancel", {}) if cancel_receipt else tender_detail
        has_merchant_receipt = "ReceiptMerchant" in container and container["ReceiptMerchant"] not in ['', None]
        has_customer_receipt = "ReceiptCustomer" in container and container["ReceiptCustomer"] not in ['', None]

        if not (has_merchant_receipt or has_customer_receipt):
            return

        try:
            model = get_model(pos_id)
            pod_type = get_podtype(model)
            current_printer = get_device_service(
                pos_id=pos_id,
                device_type="printer",
                sector=sector,
            )

            if has_merchant_receipt and self.print_merchant_receipt and pod_type != "TT":
                self._print_tender_receipt(
                    current_printer,
                    container["ReceiptMerchant"],
                    fiscalization_type=fiscalization_type
                )

            if has_customer_receipt:
                if self.ask_print_customer_receipt:
                    if print_customer_receipt is None:
                        print_customer_receipt = show_confirmation(pos_id, "$WANT_TO_PRINT_CUSTOMER_RECEIPT")
                else:
                    print_customer_receipt = self.print_customer_receipt

                if print_customer_receipt:
                    self._print_tender_receipt(
                        current_printer,
                        container["ReceiptCustomer"],
                        fiscalization_type=fiscalization_type
                    )
        except Exception:  # noqa
            self.logger.exception("Error printing TEF receipt for tender")

        return print_customer_receipt

    def _print_tender_receipt(
        self,
        current_printer,    # type: str
        receipt_data,       # type: str
        fiscalization_type=""       # type: Optional[str]
    ):
        # type: (...) -> None

        receipt = self._decode_receipt(
            receipt_data,
            fiscalization_type
        )
        self.mb_context.MB_EasySendMessage(
            current_printer,
            TK_PRN_PRINT,
            FM_PARAM,
            receipt,
            30000000
        )

    @staticmethod
    def _decode_receipt(
        receipt_data,           # type: str
        fiscalization_type      # type: str
    ):
        # type: (...) -> str

        append_cut_command = not fiscalization_type or not fiscalization_type.startswith("APOS")
        cut_tag = b"{}SpaceCutReceiptCard{}"
        space_cut_receipt_card = cut_tag.format(SeparatorTag.start, SeparatorTag.end) if append_cut_command else ""
        receipt = "{}{}".format(
            base64.b64decode(receipt_data),
            space_cut_receipt_card
        )
        return receipt



from typing import Any, Optional
from messagebus import MessageBus, Message, Event
from abc import ABCMeta, abstractmethod


class MessageProcessor(object):
    """The methods of this interface are called based on the lifecycle of the default message handling process"""
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_processor_name(self):
        # type: () -> str
        """
        This method is called to inform callbacks the name of the processor that is being executed
        :return: the name of the processor that is being processed
        """
        raise NotImplementedError()

    @abstractmethod
    def generate_unique_id(self):
        # type: () -> str
        """
        This method is called so that the Processor can generate an unique id for the processing. The value generated
        by this method is passed to callbacks to inform about the request that is being processed
        :return: the unique id generated by the processor
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_data(self, data, event_type):
        # type: (bytes, Optional[str]) -> Any
        """
        This is the first method called. It received the message data and is responsible for parsing it and returning
        the parsed object. In case of parsing errors, an exception must be raised and it must he handled in the
        format_parse_exception method
        :param data: the data of the message
        :param event_type: if handling an event and this event has a type, this will be the type of the event
        :return: the parsed object
        """
        raise NotImplementedError()

    @abstractmethod
    def call_business(self, obj):
        # type: (Any) -> Any
        """
        This method is responsible for executing the request. It should call business objects with the parsed parameter
        and return the value to be formatted
        :param obj: the object parsed by the parse_data method
        :return: the result of the operation
        """
        raise NotImplementedError()

    @abstractmethod
    def format_response(self, message_bus, message, event, input_obj, result):
        # type: (MessageBus, Optional[Message], Optional[Event], Any, Any) -> None
        """
        This method is responsible for formatting a response. It receives the object parsed by the parse_data and the
        result returned by the call_business and must interact with the message bus to produce a response
        :param message_bus: the interface of the message bus
        :param message: the message that is being processed or None if an event is being processed
        :param event:  the event that is being processed or None if a message is being processed
        :param input_obj: the object that was parsed by the parse_data method
        :param result: the value returned by the call_business
        """
        raise NotImplementedError()

    def format_exception(self, message_bus, message, event, input_obj, exception):
        # type: (MessageBus, Optional[Message], Optional[Event], Any, BaseException) -> None
        """
        This method is responsible for formatting a response in case of any exception produced by the previous steps
        after the data ir parsed.
        :param message_bus: the interface to the message bus
        :param message: the message that is being processed or None if an event is being processed
        :param event:  the event that is being processed or None if a message is being processed
        :param input_obj: the object that was parsed by the parse_data method
        :param exception: the exception produced by validate_data or call_business methods
        """
        raise NotImplementedError()

    def format_parse_exception(self, message_bus, message, event, data, exception):
        # type: (MessageBus, Optional[Message], Optional[Event], bytes, BaseException) -> None
        """
        This method is responsible for formatting a response in case of any exception produced by the parse_data method.
        :param message_bus: the interface to the message bus
        :param message: the message that is being processed or None if an event is being processed
        :param event:  the event that is being processed or None if a message is being processed
        :param data: the data received by the parse_data method
        :param exception: the exception produced by validate_data or call_business methods
        """
        raise NotImplementedError()



import sysactions

ask_document_pinpad = {}


def get_ask_document_pinpad(
    pos_id,  # type: str
):
    # type: (...) -> bool

    global ask_document_pinpad

    if pos_id not in ask_document_pinpad:
        ask_document_pinpad[str(pos_id)] = sysactions.get_cfg(pos_id).find_value("AskDocumentPinpad", "false").lower() == "true"

    return ask_document_pinpad[pos_id]



# -*- coding: utf-8 -*-
import persistence
import pyscripts
import copy
import logging
import json

from xml.etree.ElementTree import Element
from xml.etree import cElementTree as eTree
from cache import CacheManager, InfiniteExpiration
from typing import (
    List,
    Optional
)
from helper import get_valid_pricelists
from msgbus import TK_SYS_ACK
from bustoken import TK_REMOTE_ORDER_GET_DELIVERY_BRANDS

logger = logging.getLogger("PosActions")


class ModifierSets:
    def __init__(self):
        self.context = ""
        self.level = -1
        self.line_number = -1
        self.modifiers_sets = {}

    def to_xml(self):
        element = eTree.Element("ModifierSets")  # type: Element

        element.set("itemId", self.context.decode("UTF-8"))
        element.set("level", str(self.level))
        element.set("lineNumber", str(self.line_number))

        for modifier_key in self.modifiers_sets:
            element.append(self.modifiers_sets[modifier_key].to_xml())

        return element


class ModifierSet:
    def __init__(self):
        self.context = ""
        self.part_code = ""
        self.product_name = ""
        self.display_name = ""
        self.default_qty = ""
        self.min_qty = ""
        self.max_qty = ""
        self.min_item_qty = ""
        self.max_item_qty = ""
        self.json_array_tags = ""
        self.modifiers = {}

    def to_xml(self):
        element = eTree.Element("ModifierSet")  # type: Element

        element.set("defaultQty", str(self.default_qty))
        element.set("includedQty", str(self.default_qty))
        element.set("partCode", self.part_code)

        element.set("productName", self.product_name.decode("UTF-8"))
        element.set("displayName", self.display_name.decode("UTF-8"))
        element.set("jsonArrayTags", str(self.json_array_tags))

        element.set("minQty", str(self.min_qty))
        element.set("maxQty", str(self.max_qty))
        element.set("minItemQty", str(self.min_item_qty))
        element.set("maxItemQty", str(self.max_item_qty))

        for modifier_key in self.modifiers:
            element.append(self.modifiers[modifier_key].to_xml())

        return element


class Modifier:
    def __init__(self):
        self.context = ""
        self.part_code = ""
        self.product_name = ""
        self.unit_price = ""
        self.add_price = ""
        self.sub_price = ""
        self.color = ""
        self.def_qty = 0
        self.ord_qty = 0
        self.level = -1
        self.line_number = -1
        self.json_array_tags = ""
        self.parts = []

    def to_xml(self):
        element = eTree.Element("Modifier")  # type: Element

        element.set("itemId", self.context)
        element.set("partCode", self.part_code)

        element.set("productName", self.product_name.decode("UTF-8"))
        element.set("color", self.color)
        element.set("jsonArrayTags", str(self.json_array_tags))

        element.set("addedUnitPrice", str(self.add_price))
        element.set("subUnitPrice", str(self.sub_price))
        element.set("itemPrice", str(self.unit_price))

        element.set("defaultQty", str(self.def_qty))
        element.set("orderedQty", str(self.ord_qty))

        element.set("level", str(self.level))
        element.set("lineNumber", str(self.line_number))

        for part_key in self.parts:
            element.append(self.parts[part_key].to_xml())

        return element


class Part:
    def __init__(self):
        self.context = ""
        self.part_code = ""
        self.product_name = ""
        self.unit_price = ""
        self.add_price = ""
        self.sub_price = ""
        self.def_qty = 0
        self.min_qty = 0
        self.max_qty = 0
        self.level = -1
        self.line_number = -1
        self.json_array_tags = ""

    def to_xml(self):
        element = eTree.Element("Part")  # type: Element

        element.set("itemId", self.context)
        element.set("partCode", self.part_code)
        element.set("productName", self.product_name)
        element.set("jsonArrayTags", str(self.json_array_tags))

        element.set("addedUnitPrice", str(self.add_price))
        element.set("subUnitPrice", str(self.sub_price))
        element.set("itemPrice", str(self.unit_price))

        element.set("defQty", str(self.def_qty))
        element.set("minQty", str(self.min_qty))
        element.set("maxQty", str(self.max_qty))

        element.set("level", str(self.level))
        element.set("lineNumber", str(self.line_number))

        return element


class Cache:
    def __init__(self, mbcontext):
        self.mbcontext = mbcontext
        self._delivery_brands_cache = CacheManager(
            InfiniteExpiration(),
            new_object_func=self._delivery_brands_renewer
        )
        self._tags_cache = CacheManager(InfiniteExpiration(), new_object_func=self._tags_cache_renewer)
        self._parts_cache = CacheManager(InfiniteExpiration(), new_object_func=self._parts_cache_renewer)
        self._product_prices_cache = CacheManager(InfiniteExpiration(), new_object_func=self._product_prices_cache_renewer)
        self._price_key_cache = CacheManager(InfiniteExpiration(), new_object_func=self._price_key_cache_renewer)
        self._must_modify_cache = CacheManager(InfiniteExpiration(), new_object_func=self._must_modify_cache_renewer)
        self._dt_only_cache = CacheManager(InfiniteExpiration(), new_object_func=self._dt_only_cache_renewer)
        self._discount_cache = CacheManager(InfiniteExpiration(), new_object_func=self._discount_cache_renewer)
        self._options_cache = CacheManager(InfiniteExpiration(), new_object_func=self._options_cache_renewer)

    def seach_items_tree_combo(self, part_code, list_categories=[]):
        """
        Busca na arvore de um combo e retorna o ultimo item da mesma
        :param part_code: product_code ou product_part
        :return: ultimo id da arvore
        """
        tags_cache = self._tags_cache.get_cached_object()
        p_code = [part_code] if len(self.get_parts(part_code)) is 0 else filter(lambda x: x, self.get_parts(part_code))

        if len(p_code) > 1:
            list_items = []

            for code in p_code:
                if len(self.get_parts(code)) is 0:
                    tag_kiosk = tags_cache[code][0]
                    list_items.append(1 if (tag_kiosk.get('CATEGORY', None) in list_categories) else 0)
            return list_items
        else:
            if len(self.get_parts(p_code[0])) is 0:
                tag_kiosk = tags_cache[p_code[0]][0]
                return [1 if (tag_kiosk.get('CATEGORY', None) in list_categories) else 0]

        return self.seach_items_tree_combo(part_code, list_categories)
    # END of seach_items_tree_combo

    def is_not_order_kiosk(self, part_code, podtype, list_categories=[]):
        """
        Verifica se o pedido tem a tag de sobremesas para ser vendido, apenas valido para o POS 8
        :param part_code: id do produto
        :param podtype: Point-of-distribution type
        :return: Boolean
        """
        if podtype == 'KK':
            parts = [part_code] if len(self.get_parts(part_code)) is 0 else self.get_parts(part_code)
            last_items_tree = [self.seach_items_tree_combo(p_code, list_categories) for p_code in parts]
            result = list(set(sum(last_items_tree, [])))

            if not len(result) > 1:
                return True if result[0] is 0 else False
            return True
    # END of is_not_order_kiosk

    def get_tags_as_dict(self, part_code, wanted_tag):
        # type: (str, str) -> str

        tags_cache = self._tags_cache.get_cached_object()

        if wanted_tag:
            wanted_tag = wanted_tag.upper()
            if part_code in tags_cache and wanted_tag in tags_cache[part_code][0]:
                return tags_cache[part_code][0][wanted_tag]
        else:
            if part_code in tags_cache:
                return tags_cache[part_code][0]

        return ''
    # END of get_tags

    def get_tags_as_str(self, part_code):
        # type: (str) -> str

        tags_cache = self._tags_cache.get_cached_object()
        if part_code in tags_cache:
            return tags_cache[part_code][1]

        return ''
    # END of get_tags

    def get_parts(self, part_code):
        # type: (str) -> dict

        parts_cache = self._parts_cache.get_cached_object()
        if part_code in parts_cache:
            return parts_cache[part_code]

        return {}
    # END of get_parts

    def get_options(self, part_code):
        # type: (str) -> Optional[ModifierSet]
        options_cache = self._options_cache.get_cached_object()

        if part_code in options_cache:
            return copy.deepcopy(options_cache[part_code])

        return None
    # END of get_options

    def get_best_price(self, context, part_code, pod_type="EI", forced_pricelist=''):
        # type: (str, str, str, str) -> Optional[dict]

        best_price, _ = self._get_best_price_and_key(context, part_code, pod_type, forced_pricelist=forced_pricelist)
        return best_price

    def get_best_price_key(self, context, part_code, pod_type="EI", forced_pricelist=''):
        # type: (str, str, str, str) -> Optional[dict]

        _, price_key = self._get_best_price_and_key(context, part_code, pod_type, forced_pricelist=forced_pricelist)
        return price_key

    def _get_best_price_and_key(self, context, part_code, pod_type, forced_pricelist=''):
        best_price = None
        best_price_key = None

        if pod_type == "DL":
            best_price, best_price_key = self._find_best_price_and_key_by_pod_type(
                context, part_code, "DL", forced_pricelist=forced_pricelist
            )
        if best_price is None and best_price_key is None:
            best_price, best_price_key = self._find_best_price_and_key_by_pod_type(
                context, part_code, "EI", forced_pricelist=forced_pricelist
            )

        return best_price, best_price_key

    def _find_best_price_and_key_by_pod_type(self, current_context, part_code, pod_type, forced_pricelist=''):
        all_contexts = Cache._get_all_product_contexts(current_context)

        for context in all_contexts:
            price = self._get_price_in_context(context, part_code, pod_type, forced_pricelist=forced_pricelist)
            price_key = self._get_price_key_in_context(context, part_code, pod_type, forced_pricelist=forced_pricelist)
            if price is not None:
                return price, price_key
        return None, None

    @staticmethod
    def _get_all_product_contexts(current_context):
        contexts = []
        index = 0

        while index >= 0:
            contexts.append(current_context)
            index = current_context.find('.')
            current_context = current_context[index + 1:]

        contexts.append("")
        return contexts

    def _get_price_in_context(self, context, part_code, pod_type, forced_pricelist=''):
        key = (part_code, context)
        if forced_pricelist:
            pod_type = forced_pricelist
        pod_types = pod_type.split('.')[::-1]
        for price_list in pod_types:
            if price_list not in self._product_prices_cache.get_cached_object().keys():
                continue
            product_price_cache = self._product_prices_cache.get_cached_object()[price_list]
            price = product_price_cache[key] if key in product_price_cache else None
            if price:
                return price
        return None

    def _get_price_key_in_context(self, context, part_code, pod_type, forced_pricelist=''):
        key = (part_code, context)
        if forced_pricelist:
            pod_type = forced_pricelist
        pod_types = pod_type.split('.')[::-1]
        for price_list in pod_types:
            if price_list not in self._price_key_cache.get_cached_object().keys():
                continue
            product_price_key_cache = self._price_key_cache.get_cached_object()[price_list]
            price_key = product_price_key_cache[key] if key in product_price_key_cache else None
            if price_key:
                return price_key
        return None

    def get_must_modify(self, part_code, pod_type):
        must_modify_cache = self._must_modify_cache.get_cached_object()
        dt_only_cache = self._dt_only_cache.get_cached_object()

        if part_code in must_modify_cache:
            if pod_type != "DT" and part_code in dt_only_cache:
                return None
            else:
                return must_modify_cache[part_code]

        return None
    # END get_best_price_key

    def get_discount(self, part_code, discount_code):
        discount_cache = self._discount_cache.get_cached_object()
        key = (part_code, discount_code)
        if key in discount_cache:
            return discount_cache[key]

        return None
    # END get_best_price_key

    def get_delivery_brands(self):
        # type: () -> List

        return self._delivery_brands_cache.get_cached_object()

    def _delivery_brands_renewer(self):
        # type: () -> List

        try:
            result = self.mbcontext.MB_EasySendMessage(
                dest_name="RemoteOrder",
                token=TK_REMOTE_ORDER_GET_DELIVERY_BRANDS,
                data=""
            )
            if result.token == TK_SYS_ACK:
                return json.loads(result.data)

            return []

        except Exception as _:
            return []

    def _tags_cache_renewer(self):
        conn = None
        try:
            conn = persistence.Driver().open(self.mbcontext)
            cursor = conn.select("""SELECT ProductCode, upper(Tag) FROM productdb.ProductTags ORDER BY ProductCode""")
            temp_cache = [(x.get_entry(0), x.get_entry(1)) for x in cursor]

            if not temp_cache:
                return {}

            tags_cache = {}
            tags = {}
            tag_str = ""
            current_code = temp_cache[0][0]
            for tags_line in temp_cache:
                if current_code != tags_line[0]:
                    tags_cache[current_code] = (tags, tag_str)
                    current_code = tags_line[0]
                    tags = {}
                    tag_str = ""

                tmp = tags_line[1].split('=')
                tags[tmp[0].upper()] = tmp[1] if len(tmp) > 1 else ''
                tag_str = tag_str + tags_line[1]

            tags_cache[current_code] = (tags, tag_str)
            return tags_cache
        except (persistence.AprException, persistence.DbdException) as _:
            logger.exception('Error getting tags to renew cache')
            return {}
        finally:
            if conn:
                conn.close()

    def _parts_cache_renewer(self):
        conn = None
        try:
            conn = persistence.Driver().open(self.mbcontext)
            query = """SELECT  ProductPart.ProductCode, 
                                ProductPart.PartCode, 
                                DefaultQty, 
                                MinQty, 
                                MaxQty, 
                                ProductName 
                         FROM ProductPart 
                        INNER JOIN Product ON ProductPart.PartCode = Product.ProductCode 
                        WHERE PartType != 1 
                        ORDER BY ProductPart.ProductCode"""
            cursor = conn.select(query)
            temp_cache = [
                (
                    x.get_entry(0),
                    x.get_entry(1),
                    (
                        x.get_entry(2),
                        x.get_entry(3),
                        x.get_entry(4),
                        x.get_entry(5)
                     )
                ) for x in cursor
            ]

            if not temp_cache:
                return {}

            parts_cache = {}
            parts = {}
            current_code = temp_cache[0][0]
            for part_line in temp_cache:
                if current_code != part_line[0]:
                    parts_cache[current_code] = parts
                    current_code = part_line[0]
                    parts = {}

                parts[part_line[1]] = part_line[2]

            parts_cache[current_code] = parts
            return parts_cache
        except (persistence.AprException, persistence.DbdException) as _:
            logger.exception('Error getting parts to renew cache')
            return {}
        finally:
            if conn:
                conn.close()

    def _product_prices_cache_renewer(self):
        conn = None
        try:
            lists = get_valid_pricelists(self.mbcontext)
            all_prices = {}
            conn = persistence.Driver().open(self.mbcontext)
            for price_list in lists:
                all_prices[price_list] = Cache._query_all_prices_by_price_list(conn, price_list)

            return all_prices
        finally:
            if conn:
                conn.close()

    @staticmethod
    def _query_all_prices_by_price_list(conn, price_list):
        query_all_price_list = """
            SELECT productcode, 
                   context, 
                   defaultunitprice, 
                   addedunitprice, 
                   subtractedunitprice, 
                   pricelistid 
            FROM   price 
            WHERE  validthru >= Datetime(Datetime(), 'localtime') 
                   AND validfrom <= Datetime(Datetime(), 'localtime') 
                   AND pricelistid = '{}' """

        return {(str(x.get_entry(0)), str(x.get_entry(1) or '')):
                (x.get_entry(2) or '0.00', x.get_entry(3) or '0.00', x.get_entry(4) or 0)
                for x in conn.select(query_all_price_list.format(price_list))}

    def _price_key_cache_renewer(self):
        conn = None
        try:
            lists = get_valid_pricelists(self.mbcontext)
            all_prices_key = {}
            conn = persistence.Driver().open(self.mbcontext)
            for price_list in lists:
                all_prices_key[price_list] = Cache._query_all_price_keys_by_price_list(conn, price_list)

            return all_prices_key
        finally:
            if conn:
                conn.close()

    @staticmethod
    def _query_all_price_keys_by_price_list(conn, price_list):
        query_all_price_keys_by_price_list = """
            SELECT productcode, 
                   context, 
                   pricekey, 
                   pricelistid 
            FROM   price 
            WHERE  validthru >= Datetime(Datetime(), 'localtime') 
                   AND validfrom <= Datetime(Datetime(), 'localtime') 
                   AND pricelistid = '{}' """

        return {(str(x.get_entry(0)), str(x.get_entry(1) or '')): x.get_entry(2)
                for x in conn.select(query_all_price_keys_by_price_list.format(price_list))}

    def _must_modify_cache_renewer(self):
        conn = None
        try:
            conn = persistence.Driver().open(self.mbcontext)
            return {str(x.get_entry(0)): str(x.get_entry(1)).split("=")[1] for x in conn.select(
                """SELECT ProductCode, Tag from ProductTags where Tag like 'MustBeModifiedSet%' """)}
        finally:
            if conn:
                conn.close()

    def _dt_only_cache_renewer(self):
        conn = None
        try:
            conn = persistence.Driver().open(self.mbcontext)
            return {str(x.get_entry(0)): str(x.get_entry(1)).split("=")[1] for x in
                    conn.select("""SELECT ProductCode, Tag from ProductTags where Tag like 'DTOnly%' """)}
        finally:
            if conn:
                conn.close()

    def _discount_cache_renewer(self):
        conn = None
        try:
            conn = persistence.Driver().open(self.mbcontext)
            return {(str(x.get_entry(0)), str(x.get_entry(1))): (x.get_entry(3), x.get_entry(4)) for x
                    in conn.select("""SELECT * from ProductDiscounts where Exclude == 'N'""")}
        finally:
            if conn:
                conn.close()

    def _options_cache_renewer(self):
        conn = None
        try:
            conn = persistence.Driver().open(self.mbcontext)
            query_result = [(str(x.get_entry(0)), str(x.get_entry(1)), str(x.get_entry(2)), str(x.get_entry(3)),
                             str(x.get_entry(4)), str(x.get_entry(5))) for x in conn.select("""
            select pc.ClassCode, p.ProductName, pt.Tag, pc.ProductCode, p2.ProductName, pt2.Tag from ProductClassification pc
            inner join Product p
            on pc.ClassCode = p.ProductCode
            left outer join ProductTags pt
            on p.ProductCode = pt.ProductCode
            inner join Product p2
            on p2.ProductCode = pc.ProductCode
            left outer join ProductTags pt2
            on p2.ProductCode = pt2.ProductCode
            where (pt.Tag like 'DisplayName%%' or pt.Tag like 'ItemQtys%%' or pt2.Tag like 'Color%%')
            and ClassCode <> 1
            order by pc.ClassCode, p.ProductCode""")]

            if not query_result:
                return

            new_option_cache = {}

            current_modifier_set = ModifierSet()
            current_modifier_set.part_code = query_result[0][0]
            current_modifier_set.product_name = query_result[0][1]
            current_modifier_set.default_qty = 1
            current_modifier_set.min_qty = 1
            current_modifier_set.max_qty = 1
            current_modifier_set.min_item_qty = 0
            current_modifier_set.max_item_qty = 1
            if query_result[0][2].upper().startswith("DISPLAYNAME"):
                current_modifier_set.display_name = query_result[0][2].split("=")[1]
            elif query_result[0][2].upper().startswith("ITEMQTYS"):
                item_qtys = query_result[0][2].split("=")[1]
                current_modifier_set.min_item_qty = item_qtys.split(";")[0]
                current_modifier_set.max_item_qty = item_qtys.split(";")[1]

            products_processed = {}
            modifiers = {}
            for query_line in query_result:
                if current_modifier_set.part_code != query_line[0]:
                    current_modifier_set.modifiers = modifiers
                    new_option_cache[current_modifier_set.part_code] = current_modifier_set

                    current_modifier_set = ModifierSet()
                    current_modifier_set.part_code = query_line[0]
                    current_modifier_set.product_name = query_line[1]
                    current_modifier_set.default_qty = 1
                    current_modifier_set.min_qty = 1
                    current_modifier_set.max_qty = 1
                    current_modifier_set.min_item_qty = 0
                    current_modifier_set.max_item_qty = 1

                    modifiers = {}
                    products_processed = {}

                if query_line[2].upper().startswith("DISPLAYNAME"):
                    current_modifier_set.display_name = query_line[2].split("=")[1]
                elif query_line[2].upper().startswith("ITEMQTYS"):
                    item_qtys = query_line[2].split("=")[1]
                    current_modifier_set.min_item_qty = item_qtys.split(";")[0]
                    current_modifier_set.max_item_qty = item_qtys.split(";")[1]

                if query_line[3] not in products_processed:
                    mod = Modifier()
                    mod.part_code = query_line[3]
                    mod.product_name = query_line[4]
                    if query_line[5].upper().startswith("COLOR"):
                        mod.color = query_line[5].split("=")[1]

                    modifiers[mod.part_code] = mod
                    products_processed[mod.part_code] = mod
                else:
                    if query_line[5].upper().startswith("COLOR"):
                        mod = products_processed[query_line[3]]
                        mod.color = query_line[5].split("=")[1]

            current_modifier_set.modifiers = modifiers
            new_option_cache[current_modifier_set.part_code] = current_modifier_set

            return new_option_cache
        finally:
            if conn:
                conn.close()


cache = Cache(pyscripts.mbcontext)



from Tkinter import *
import ttk
import tkFileDialog


def main():
    root = create_root()
    frame = create_frame(root)

    # ttk.Button(root, text="Hello World").grid()
    tree = create_treeview(frame)

    text = Text(frame, wrap='none')
    text.grid(column=0, row=2, sticky=(N, W, E, S), pady=12)

    v_scroll_bar = ttk.Scrollbar(frame, orient=VERTICAL, command=text.yview)
    set_mousewheel(text, lambda e: text.yview('scroll', e.delta < 0 and 1 or -1, 'units'))

    v_scroll_bar.grid(column=1, row=2, sticky=(N, W, E, S), pady=12)
    text['yscrollcommand'] = v_scroll_bar.set

    h_scroll_bar = ttk.Scrollbar(frame, orient=HORIZONTAL, command=text.xview)
    set_mousewheel(text, lambda e: text.xview('scroll', e.delta < 0 and 1 or -1, 'units'))

    h_scroll_bar.grid(column=0, row=3, sticky=(N, W, E, S))
    text['xscrollcommand'] = h_scroll_bar.set
    for i in range(1, 40):
        tree.insert("", i, i, text=str(i), values=('20200101', 'POS_STATECHANGED', '123456789', '2020-01-01T00:00:00.000', '2020-01-01T00:00:00.000'))

    root.mainloop()


def create_treeview(frame):
    tree = ttk.Treeview(frame, columns=('business_period', 'type', 'order_id', 'creation_date', 'event_date'))
    tree.grid(column=0, row=1, sticky=(N, W, E, S))
    tree.column('#0', width=50)
    tree.heading('#0', text='ID')
    tree.column('business_period', width=100)
    tree.heading('business_period', text='Business Period')
    tree.column('type', width=200)
    tree.heading('type', text='Event Type')
    tree.column('order_id', width=100)
    tree.heading('order_id', text='Order ID')
    tree.column('creation_date', width=150)
    tree.heading('creation_date', text='Creation Date')
    tree.column('event_date', width=150)
    tree.heading('event_date', text='Event Date')

    scroll_bar = ttk.Scrollbar(frame, orient=VERTICAL, command=tree.yview)
    set_mousewheel(tree, lambda e: tree.yview('scroll', e.delta < 0 and 1 or -1, 'units'))

    scroll_bar.grid(column=1, row=1, sticky=(N, W, E, S))
    tree['yscrollcommand'] = scroll_bar.set
    return tree


def create_frame(root):
    frame = ttk.Frame(root, padding="12 12 12 12")
    frame.grid(column=0, row=0, sticky=(N, W, E, S))
    frame.columnconfigure(0, weight=1)
    frame.rowconfigure(0, weight=1)
    frame.rowconfigure(1, weight=1)
    return frame


def create_root():
    root = Tk()
    root.columnconfigure(0, weight=1)
    root.rowconfigure(0, weight=1)
    return root


def set_mousewheel(widget, command):
    """Active / deactivate mousewheel scrolling when
    cursor is over / not over the widget respectively."""
    widget.bind("<Enter>", lambda _: widget.bind_all('<MouseWheel>', lambda e: command(e)))
    widget.bind("<Leave>", lambda e: widget.unbind_all('<MouseWheel>'))


if __name__ == '__main__':
    main()



from application.interactor import GetLoyaltyIdInteractor, GetAndLockVoucherInteractor, BurnVoucherInteractor, \
    UnlockVoucherInteractor, LoyaltyAvailableInteractor
from application.model import ListenedTokens
from application.model.configuration import Configurations
from application.processor import GetLoyaltyIdProcessor, GetAndLockVoucherProcessor, BurnVoucherProcessor, \
    UnlockVoucherProcessor, LoyaltyAvailableProcessor
from application.repository.api import APIRepository
from messagehandler import MessageHandlerBuilder
from messageprocessor import \
    MessageProcessorMessageHandler, \
    DefaultMessageProcessorExecutorFactory, \
    LoggingProcessorCallback
from typing import Union


class ComponentMessageHandlerBuilder(MessageHandlerBuilder):

    def __init__(self, configs):
        # type: (Configurations) -> None

        self.logger = configs.logger
        self.configs = configs
        self.api_repository = APIRepository(self.configs)

        self.message_handler = None  # type: Union[None, MessageProcessorMessageHandler]

    def build_singletons(self):
        # type: () -> None

        get_loyalty_id_interactor = GetLoyaltyIdInteractor(self.configs, self.api_repository)
        get_and_lock_voucher_interactor = GetAndLockVoucherInteractor(self.configs, self.api_repository)
        burn_voucher_interactor = BurnVoucherInteractor(self.configs, self.api_repository)
        unlock_voucher_interactor = UnlockVoucherInteractor(self.configs, self.api_repository)
        loyalty_available_interactor = LoyaltyAvailableInteractor(self.configs)

        get_loyalty_id_processor = GetLoyaltyIdProcessor(self.configs, get_loyalty_id_interactor)
        get_and_lock_voucher_processor = GetAndLockVoucherProcessor(self.configs, get_and_lock_voucher_interactor)
        burn_voucher_processor = BurnVoucherProcessor(self.configs, burn_voucher_interactor)
        unlock_voucher_processor = UnlockVoucherProcessor(self.configs, unlock_voucher_interactor)
        loyalty_available_processor = LoyaltyAvailableProcessor(loyalty_available_interactor)

        messages_processors = {
            ListenedTokens.TK_LOYALTY_GET_LOYALTY_ID.value: get_loyalty_id_processor,
            ListenedTokens.TK_LOYALTY_GET_AND_LOCK_VOUCHER.value: get_and_lock_voucher_processor,
            ListenedTokens.TK_LOYALTY_BURN_VOUCHER.value: burn_voucher_processor,
            ListenedTokens.TK_LOYALTY_UNLOCK_VOUCHER.value: unlock_voucher_processor,
            ListenedTokens.TK_LOYALTY_IS_AVAILABLE.value: loyalty_available_processor
        }

        callbacks = [LoggingProcessorCallback(self.logger)]

        self.message_handler = MessageProcessorMessageHandler(
            None,
            messages_processors,
            None,
            DefaultMessageProcessorExecutorFactory(callbacks),
            logger=self.logger
        )

    def build_message_handler(self):
        # type: () -> None

        return self.message_handler

    def destroy_singletons(self):
        # type: () -> None

        return

    def destroy_message_handler(self, message_handler):
        # type: (MessageHandlerBuilder) -> None

        return



import os
import sys

import helper

from mwposreport import ReporterFactory
from old_helper import (
    get_class_by_name,
    config_logger
)
import reports


def main():
    helper.import_pydevd(os.environ["LOADERCFG"], 9124)
    current_module = sys.modules[__name__]
    config_logger(os.environ["LOADERCFG"], 'ReportActions')

    available_reports = []
    for _, dir_names, _ in os.walk(os.path.join(os.path.dirname(__file__), "reportsentry")):
        available_reports = dir_names
        break

    for report in available_reports:
        setattr(current_module, report, create_wrapper(report, _custom_report_impl))


def _custom_report_impl(report_name, pos_id, *args, **kwargs):
    complete_class_name = "reportsentry." + report_name + ".ReportGenerator"

    report_class = get_class_by_name(unicode(complete_class_name))

    report = getattr(report_class, "generate")(reports.mbcontext, pos_id, *args, **kwargs)
    if isinstance(report, tuple):
        generator = report[0]
        presenter = report[1]
        report_data = ReporterFactory.build_reporter(pos_id, generator, presenter).create_report()
        return report_data
    else:
        return report


def create_wrapper(name, func):
    def wrapper(*args, **kwargs):
        return func(name, args[0], *args[1:], **kwargs)

    return wrapper


main()



import datetime

from typing import Optional, Union


class Line(object):
    def __init__(self,
                 order_id,
                 operator,
                 authorizator,
                 line_number,
                 quantity,
                 part_code,
                 price_key,
                 void_datetime,
                 price=None,
                 operator_name=""):
        # type: (int, int, Union[int, None], int, int, int, int, datetime, Optional[float], Optional[str]) -> None
        self.order_id = order_id
        self.operator = operator
        self.authorizator = authorizator
        self.line_number = line_number
        self.quantity = quantity
        self.part_code = part_code
        self.price_key = price_key
        self.void_datetime = void_datetime
        self.price = price
        self.operator_name = operator_name



# coding=utf-8
import json
from logging import Logger

from helper import json_serialize, remove_accents
from requests import Response


class ConfigurationReaderException(Exception):
    pass


class NotFound(Exception):
    pass


class ValidationError(Exception):
    pass


class ResponseError(Exception):
    def __init__(self, logger, response):
        # type: (Logger, Response) -> None

        self.logger = logger
        self.response = response
        self.error_status_code = response.status_code
        self.message = _get_response_content(response)
        response_str = json_serialize(response)

        message = "Unexpected response received. StatusCode: {}; Content: {}; FullResponse:\n{}"
        self.logger.error(message.format(self.error_status_code, self.message, response_str))

        super(ResponseError, self).__init__(self.message)


class NoResponseError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "No response. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(NoResponseError, self).__init__(content)


class ForbiddenError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Forbidden error. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(ForbiddenError, self).__init__(content)


class NotFoundError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Not found. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(NotFoundError, self).__init__(content)


class NotAvailableError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Code not available. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(NotAvailableError, self).__init__(content)


class InUseError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Code in use on another terminal. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(InUseError, self).__init__(content)


class AlreadyUnlockedError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Code already unlocked. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(AlreadyUnlockedError, self).__init__(content)


class ProcessorError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Processing error. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(ProcessorError, self).__init__(content)


def _get_response_content(response):
    try:
        content = json.loads(response.content).get("detail").encode("utf8")
        return content
    except (AttributeError, ValueError):
        return ""


def build_message_exception(ex):
    error_type = type(ex).__name__
    error_message = ex.message
    error_detail = {"type": error_type, "message": error_message or None}
    ex.message = json.dumps(error_detail)
    return ex



from msgbus import MBEasyContext
from typing import Optional

from mw import MwTableOrderRetriever, MsgBusTableService, RealClock
from reportsdata.tablereport import TableReportPresenter, TableReportGenerator
from reportsdata.tablereport.dto import TableReportFilterDto


class ReportGenerator(object):
    @staticmethod
    def generate(
        mb_context,                         # type: MBEasyContext
        pos_id,                             # type: str
        table_id,                           # type: str
        print_pre_account_by="TABLE",       # type: Optional[str]
        hide_unpriced_items="True",         # type: Optional[str]
        pre_account_with_discount=True,     # type: Optional[bool]
        filtered_seat=0                     # type: Optional[str]
    ):
        # type: (...) -> (TableReportGenerator, TableReportPresenter)

        table_order_retriever = MwTableOrderRetriever(MsgBusTableService())

        hide_unpriced_items = True if hide_unpriced_items == "True" else False
        pre_account_with_discount = True if pre_account_with_discount == "True" else False

        generator = TableReportGenerator(
                TableReportFilterDto(pos_id, table_id, None),
                RealClock(),
                table_order_retriever
        )

        presenter = TableReportPresenter(
            print_by=print_pre_account_by,
            hide_unpriced_items=hide_unpriced_items,
            pre_account_with_discount=pre_account_with_discount,
            pos_id=pos_id,
            filtered_seat=filtered_seat
        )

        return generator, presenter



from abc import ABCMeta, abstractmethod

from typing import Dict


class ImagesServiceModel(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_images(
            self,
            images_path,        # type: str
            totem_config_path   # type: str
    ):
        # type: (...) -> Dict

        raise NotImplementedError



from abc import ABCMeta, abstractmethod

from ._HttpRequest import HttpRequest
from ._HttpResponse import HttpResponse


class HttpClient(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def send(self, request):
        # type: (HttpRequest) -> HttpResponse
        raise NotImplementedError()



import sysactions

from bustoken import TK_REMOTE_ORDER_INFORM_ORDER_READY
from msgbus import TK_SYS_ACK, FM_PARAM
from .. import mb_context


@sysactions.action
def inform_order_ready(pos_id, order_id):
    message = "$INFORM_ORDER_READY_CONFIRMATION"
    confirmation = sysactions.show_messagebox(pos_id, message, timeout=720000, buttons="$OK|$CANCEL")
    if confirmation == 1:
        return False

    msg = mb_context.MB_EasySendMessage("RemoteOrder", TK_REMOTE_ORDER_INFORM_ORDER_READY, FM_PARAM, order_id)
    if msg.token != TK_SYS_ACK:
        raise Exception("ACK not received from RemoteOrder when informing order ready " + order_id)



from httpmessageprocessorutil.automapped import auto_map
from typing import List


class ToAddItem(object):
    def __init__(
        self,
        context,        # type: str
        line_number,    # type: int
        product_code,   # type: int
        quantity        # type: int
    ):
        # type: (...) -> None

        self.context = context
        self.line_number = line_number
        self.product_code = int(product_code)
        self.quantity = int(quantity)


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str},
           "pos_id": {"name": "posId", "type": str},
           "to_add_items": {"name": "toAddItems", "type": List[ToAddItem]}})
class AddItemsRequest(object):
    def __init__(self, system_api_key, pos_id, to_add_items):
        # type: (str, str, List[ToAddItem]) -> None

        self.system_api_key = system_api_key
        self.pos_id = pos_id
        self.items = to_add_items



import logging
from typing import Dict, Optional

from messagebus import MessageBus, Event, Message, NakMessage
from messagehandler import MessageHandler

from ._DefaultMessageProcessorExecutorFactory import DefaultMessageProcessorExecutorFactory
from ._MessageProcessor import MessageProcessor
from ._MessageProcessorExecutorFactory import MessageProcessorExecutorFactory

logger = logging.getLogger("MessageProcessor.MessageProcessorMessageHandler")


class MessageProcessorMessageHandler(MessageHandler):
    def __init__(self,
                 event_processors,
                 message_processors,
                 executor_factory=None,
                 subject_counter=None):
        # type: (Optional[Dict[str, MessageProcessor]], Optional[Dict[int, MessageProcessor]], Optional[MessageProcessorExecutorFactory], Optional[Dict[str, int]]) -> None  # noqa
        self.event_processors = event_processors
        self.message_processors = message_processors
        self.executor_factory = executor_factory \
            if executor_factory is not None \
            else DefaultMessageProcessorExecutorFactory([])
        self.subject_counter = subject_counter
        self.logger = logger

    def handle_event(self, message_bus, event):
        # type: (MessageBus, Event) -> None
        try:
            if self.subject_counter is not None:
                self._count_event(event.subject)

            if self.event_processors is not None:
                if self._has_event_type(event) and self._event_type_is_in_event_processors(event):
                    self._create_executor_and_execute(message_bus, event, self._get_event_type_key(event))
                    return

                if self._event_subject_is_in_event_processors(event):
                    self._create_executor_and_execute(message_bus, event, self._get_subject_key(event))
                    return

            raise MessageProcessorMessageHandler.EventNotMapped(event.subject, event.event_type)
        except:
            if self.logger is not None:
                self.logger.exception("Exception while handling event")
            raise

    def _count_event(self, subject):
        # type: (str) -> None
        subject_count = self.subject_counter.get(subject, 0)
        self.subject_counter[subject] = subject_count + 1

    def handle_message(self, message_bus, message):
        # type: (MessageBus, Message) -> None
        try:
            if self.message_processors is not None and message.token in self.message_processors:
                executor = self.executor_factory.build_executor(self.message_processors[message.token])
                executor.execute_processor(message_bus, message, None, message.data)
            else:
                message_bus.reply_message(message, NakMessage())
        except:
            if self.logger is not None:
                self.logger.exception("Exception handling message")
            raise

    def _has_event_type(self, event):
        return event.event_type is not None

    def _event_type_is_in_event_processors(self, event):
        return self._get_event_type_key(event) in self.event_processors

    def _get_event_type_key(self, event):
        return event.subject + "_" + event.event_type

    def _event_subject_is_in_event_processors(self, event):
        return self._get_subject_key(event) in self.event_processors

    def _get_subject_key(self, event):
        return event.subject

    def _create_executor_and_execute(self, message_bus, event, event_key):
        message_processor = self.event_processors[event_key]
        executor = self.executor_factory.build_executor(message_processor)
        executor.execute_processor(message_bus, None, event, event.data)

    class EventNotMapped(Exception):
        def __init__(self, subject, type):
            # type: (str, str) -> None
            self.subject = subject
            self.type = type

        def __str__(self):
            return "Event not mapped: {} - {}".format(self.subject, self.type)

        def __repr__(self):
            return "EventNotMapped('{}', '{}')".format(self.subject, self.type)



# -*- coding: utf-8 -*-
"""
    flask.testing
    ~~~~~~~~~~~~~

    Implements test support helpers.  This module is lazily imported
    and usually not used in production environments.

    :copyright: 2010 Pallets
    :license: BSD-3-Clause
"""
import warnings
from contextlib import contextmanager

import werkzeug.test
from click.testing import CliRunner
from werkzeug.test import Client
from werkzeug.urls import url_parse

from . import _request_ctx_stack
from .cli import ScriptInfo
from .json import dumps as json_dumps


class EnvironBuilder(werkzeug.test.EnvironBuilder):
    """An :class:`~werkzeug.test.EnvironBuilder`, that takes defaults from the
    application.

    :param app: The Flask application to configure the environment from.
    :param path: URL path being requested.
    :param base_url: Base URL where the app is being served, which
        ``path`` is relative to. If not given, built from
        :data:`PREFERRED_URL_SCHEME`, ``subdomain``,
        :data:`SERVER_NAME`, and :data:`APPLICATION_ROOT`.
    :param subdomain: Subdomain name to append to :data:`SERVER_NAME`.
    :param url_scheme: Scheme to use instead of
        :data:`PREFERRED_URL_SCHEME`.
    :param json: If given, this is serialized as JSON and passed as
        ``data``. Also defaults ``content_type`` to
        ``application/json``.
    :param args: other positional arguments passed to
        :class:`~werkzeug.test.EnvironBuilder`.
    :param kwargs: other keyword arguments passed to
        :class:`~werkzeug.test.EnvironBuilder`.
    """

    def __init__(
        self,
        app,
        path="/",
        base_url=None,
        subdomain=None,
        url_scheme=None,
        *args,
        **kwargs
    ):
        assert not (base_url or subdomain or url_scheme) or (
            base_url is not None
        ) != bool(
            subdomain or url_scheme
        ), 'Cannot pass "subdomain" or "url_scheme" with "base_url".'

        if base_url is None:
            http_host = app.config.get("SERVER_NAME") or "localhost"
            app_root = app.config["APPLICATION_ROOT"]

            if subdomain:
                http_host = "{0}.{1}".format(subdomain, http_host)

            if url_scheme is None:
                url_scheme = app.config["PREFERRED_URL_SCHEME"]

            url = url_parse(path)
            base_url = "{scheme}://{netloc}/{path}".format(
                scheme=url.scheme or url_scheme,
                netloc=url.netloc or http_host,
                path=app_root.lstrip("/"),
            )
            path = url.path

            if url.query:
                sep = b"?" if isinstance(url.query, bytes) else "?"
                path += sep + url.query

        self.app = app
        super(EnvironBuilder, self).__init__(path, base_url, *args, **kwargs)

    def json_dumps(self, obj, **kwargs):
        """Serialize ``obj`` to a JSON-formatted string.

        The serialization will be configured according to the config associated
        with this EnvironBuilder's ``app``.
        """
        kwargs.setdefault("app", self.app)
        return json_dumps(obj, **kwargs)


def make_test_environ_builder(*args, **kwargs):
    """Create a :class:`flask.testing.EnvironBuilder`.

    .. deprecated: 1.1
        Will be removed in 1.2. Construct ``flask.testing.EnvironBuilder``
        directly instead.
    """
    warnings.warn(
        DeprecationWarning(
            '"make_test_environ_builder()" is deprecated and will be removed '
            'in 1.2. Construct "flask.testing.EnvironBuilder" directly '
            "instead."
        )
    )
    return EnvironBuilder(*args, **kwargs)


class FlaskClient(Client):
    """Works like a regular Werkzeug test client but has some knowledge about
    how Flask works to defer the cleanup of the request context stack to the
    end of a ``with`` body when used in a ``with`` statement.  For general
    information about how to use this class refer to
    :class:`werkzeug.test.Client`.

    .. versionchanged:: 0.12
       `app.test_client()` includes preset default environment, which can be
       set after instantiation of the `app.test_client()` object in
       `client.environ_base`.

    Basic usage is outlined in the :ref:`testing` chapter.
    """

    preserve_context = False

    def __init__(self, *args, **kwargs):
        super(FlaskClient, self).__init__(*args, **kwargs)
        self.environ_base = {
            "REMOTE_ADDR": "127.0.0.1",
            "HTTP_USER_AGENT": "werkzeug/" + werkzeug.__version__,
        }

    @contextmanager
    def session_transaction(self, *args, **kwargs):
        """When used in combination with a ``with`` statement this opens a
        session transaction.  This can be used to modify the session that
        the test client uses.  Once the ``with`` block is left the session is
        stored back.

        ::

            with client.session_transaction() as session:
                session['value'] = 42

        Internally this is implemented by going through a temporary test
        request context and since session handling could depend on
        request variables this function accepts the same arguments as
        :meth:`~flask.Flask.test_request_context` which are directly
        passed through.
        """
        if self.cookie_jar is None:
            raise RuntimeError(
                "Session transactions only make sense with cookies enabled."
            )
        app = self.application
        environ_overrides = kwargs.setdefault("environ_overrides", {})
        self.cookie_jar.inject_wsgi(environ_overrides)
        outer_reqctx = _request_ctx_stack.top
        with app.test_request_context(*args, **kwargs) as c:
            session_interface = app.session_interface
            sess = session_interface.open_session(app, c.request)
            if sess is None:
                raise RuntimeError(
                    "Session backend did not open a session. Check the configuration"
                )

            # Since we have to open a new request context for the session
            # handling we want to make sure that we hide out own context
            # from the caller.  By pushing the original request context
            # (or None) on top of this and popping it we get exactly that
            # behavior.  It's important to not use the push and pop
            # methods of the actual request context object since that would
            # mean that cleanup handlers are called
            _request_ctx_stack.push(outer_reqctx)
            try:
                yield sess
            finally:
                _request_ctx_stack.pop()

            resp = app.response_class()
            if not session_interface.is_null_session(sess):
                session_interface.save_session(app, sess, resp)
            headers = resp.get_wsgi_headers(c.request.environ)
            self.cookie_jar.extract_wsgi(c.request.environ, headers)

    def open(self, *args, **kwargs):
        as_tuple = kwargs.pop("as_tuple", False)
        buffered = kwargs.pop("buffered", False)
        follow_redirects = kwargs.pop("follow_redirects", False)

        if (
            not kwargs
            and len(args) == 1
            and isinstance(args[0], (werkzeug.test.EnvironBuilder, dict))
        ):
            environ = self.environ_base.copy()

            if isinstance(args[0], werkzeug.test.EnvironBuilder):
                environ.update(args[0].get_environ())
            else:
                environ.update(args[0])

            environ["flask._preserve_context"] = self.preserve_context
        else:
            kwargs.setdefault("environ_overrides", {})[
                "flask._preserve_context"
            ] = self.preserve_context
            kwargs.setdefault("environ_base", self.environ_base)
            builder = EnvironBuilder(self.application, *args, **kwargs)

            try:
                environ = builder.get_environ()
            finally:
                builder.close()

        return Client.open(
            self,
            environ,
            as_tuple=as_tuple,
            buffered=buffered,
            follow_redirects=follow_redirects,
        )

    def __enter__(self):
        if self.preserve_context:
            raise RuntimeError("Cannot nest client invocations")
        self.preserve_context = True
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.preserve_context = False

        # Normally the request context is preserved until the next
        # request in the same thread comes. When the client exits we
        # want to clean up earlier. Pop request contexts until the stack
        # is empty or a non-preserved one is found.
        while True:
            top = _request_ctx_stack.top

            if top is not None and top.preserved:
                top.pop()
            else:
                break


class FlaskCliRunner(CliRunner):
    """A :class:`~click.testing.CliRunner` for testing a Flask app's
    CLI commands. Typically created using
    :meth:`~flask.Flask.test_cli_runner`. See :ref:`testing-cli`.
    """

    def __init__(self, app, **kwargs):
        self.app = app
        super(FlaskCliRunner, self).__init__(**kwargs)

    def invoke(self, cli=None, args=None, **kwargs):
        """Invokes a CLI command in an isolated environment. See
        :meth:`CliRunner.invoke <click.testing.CliRunner.invoke>` for
        full method documentation. See :ref:`testing-cli` for examples.

        If the ``obj`` argument is not given, passes an instance of
        :class:`~flask.cli.ScriptInfo` that knows how to load the Flask
        app being tested.

        :param cli: Command object to invoke. Default is the app's
            :attr:`~flask.app.Flask.cli` group.
        :param args: List of strings to invoke the command with.

        :return: a :class:`~click.testing.Result` object.
        """
        if cli is None:
            cli = self.app.cli

        if "obj" not in kwargs:
            kwargs["obj"] = ScriptInfo(create_app=lambda: self.app)

        return super(FlaskCliRunner, self).invoke(cli, args, **kwargs)



from abc import ABCMeta, abstractmethod

from report.command import Command
from ._CommandProcessor import CommandProcessor  # noqa


class CommandFactory(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_processor(self, command):
        # type: (Command) -> CommandProcessor
        raise NotImplementedError()



import json
import logging

from bustoken import (
    TK_MAINTENANCE_TERMINATE_REQUEST,
    TK_MAINTENANCE_RESTART_REQUEST,
    TK_MAINTENANCE_RECREATE_FILES_XML,
    TK_MAINTENANCE_RECREATE_FILE_NUMBER,
    TK_MAINTENANCE_CLEAN_AUDIT_LOG,
    TK_MAINTENANCE_DATABASE_CLEANUP_STATUS,
)
from events import (
    CleanDatabasesProcessor,
    MaintenanceProcessor
)
from messagehandler import EventHandler
from msgbus import (
    MBEasyContext,
    MBMessage,
    TK_SYS_ACK,
    TK_SYS_NAK,
)
from systools import sys_log_exception


logger = logging.getLogger("Maintenance")


class MaintenanceEventHandler(EventHandler):
    def __init__(
        self,
        mb_context,                 # type: MBEasyContext
        maintenance_processor,      # type: MaintenanceProcessor
        clean_databases_processor   # type: CleanDatabasesProcessor
    ):
        # type: (...) -> None

        super(MaintenanceEventHandler, self).__init__(mb_context)
        self.mb_context = mb_context
        self.maintenance_processor = maintenance_processor
        self.clean_databases_processor = clean_databases_processor

    def subscribe_events(self, subject):
        return

    def terminate_event(self):
        return

    @staticmethod
    def get_handled_tokens():
        return [
            TK_MAINTENANCE_TERMINATE_REQUEST,
            TK_MAINTENANCE_RESTART_REQUEST,
            TK_MAINTENANCE_RECREATE_FILES_XML,
            TK_MAINTENANCE_RECREATE_FILE_NUMBER,
            TK_MAINTENANCE_DATABASE_CLEANUP_STATUS
        ]

    def handle_event(
        self,
        subject,  # type: unicode
        evt_type,  # type: unicode
        data,  # type: str
        msg,  # type: MBMessage
    ):
        # type: (...) -> None

        logger.info("Event received: [{}]".format(subject))

        try:
            logger.info("Not inside maintenance window. Exiting...")
            event_data = json.loads(data)
            return
        except Exception:  # noqa
            logger.exception("error parsing event data as json")
            event_data = {}
        ignore_maintenance_window = event_data.get("ignoreMaintenanceWindow", False)
        if not ignore_maintenance_window:
            if not self.maintenance_processor.is_inside_maintenance_window():
                logger.info("Not inside maintenance window. Exiting...")
                return

        if subject == "CleanDatabaseBackups":
            self.clean_databases_processor.clean_database_backups()
        elif subject == "CleanDatabases":
            vacuum = event_data.get("vacuum", False)
            self.clean_databases_processor.clean_databases(vacuum)
        elif subject == "VacuumDatabases":
            self.clean_databases_processor.vacuum_databases()

    def handle_message(self, msg):
        try:
            if msg.token == TK_MAINTENANCE_TERMINATE_REQUEST:
                response = self.maintenance_processor.handle_terminate_request(msg)
            elif msg.token == TK_MAINTENANCE_RESTART_REQUEST:
                response = self.maintenance_processor.handle_restart_request(msg)
            elif msg.token == TK_MAINTENANCE_CLEAN_AUDIT_LOG:
                response = self.maintenance_processor.clean_audit_log()
            elif msg.token == TK_MAINTENANCE_DATABASE_CLEANUP_STATUS:
                response = self.maintenance_processor.handle_database_cleanup_status()
            else:
                response = (False, "")

            msg.token = TK_SYS_ACK if response[0] else TK_SYS_NAK
            self.mb_context.MB_ReplyMessage(msg, data=response[1])
        except Exception as ex:
            sys_log_exception("Unexpected exception handling event: {}".format(msg.token))
            msg.token = TK_SYS_NAK
            self.mb_context.MB_ReplyMessage(msg, data=str(ex))



from datetime import datetime
from typing import (
    List,
    Dict,
    Tuple,
    Any,
)

from msgbus import MBEasyContext
from persistence import Connection
from old_helper import BaseRepository

from dailygoals_services.model import (
    Order,
    OrderItem,
    GetPaidOrdersException,
    GetOrderItemsException,
)
from dailygoals_services.model.repository import OrderRepository as OrderServicesRepository


class OrderRepository(
    OrderServicesRepository,    # type: OrderServicesRepository
    BaseRepository,             # type: BaseRepository
):

    def __init__(
        self,
        mb_context,     # type: MBEasyContext
        pos_list,       # type: List
    ):
        # type: (...) -> None

        super(OrderRepository, self).__init__(mb_context)
        self.pos_list = pos_list

    def get_paid_ticket(
        self,
        business_period,    # type: str
    ):
        # type: (...) -> List

        def inner_func(
            conn,   # type: Connection
        ):
            # type: (...) -> List[Tuple[Any, Any]]

            query = self._PaidTicketsQuery.format(business_period)
            users_average_ticket = [(x.get_entry(0), x.get_entry(1)) for x in conn.select(sql=query)]
            return users_average_ticket

        return self.execute_parallel_in_all_databases_returning_flat_list(
            method=inner_func,
            pos_list=self.pos_list,
        )

    def get_paid_orders_ticket(
        self,
        business_period,    # type: str
    ):
        # type: (...) -> List

        def inner_func(
            conn,   # type: Connection
        ):
            # type: (...) -> List

            query = self._PaidOrdersTicketQuery.format(business_period)
            orders = [[(x.get_entry(0), x.get_entry(1)) for x in conn.select(sql=query)][0]]
            return orders

        return self.execute_parallel_in_all_databases_returning_flat_list(
            method=inner_func,
            pos_list=self.pos_list,
        )

    def get_paid_orders(
        self,
        operator_id,    # type: str
        initial_date,   # type: datetime
        end_date,       # type: datetime
    ):
        # type: (...) -> List

        try:
            def inner_func(
                conn,   # type: Connection
            ):
                # type: (...) -> List

                query = self._PaidOrdersQuery
                date_format = "%Y%m%d"

                query = query.format(
                    initial_date.strftime(date_format),
                    end_date.strftime(date_format),
                    int(operator_id)
                )
                if operator_id in ["", "0", "None"]:
                    query = query.replace("user={},".format(int(operator_id)), "user=")

                all_orders = {}  # type: Dict[int, Order]
                order_with_payments = [(
                    int(x.get_entry(0)),
                    int(x.get_entry(1)),
                    datetime.strptime(x.get_entry(2), "%Y%m%d"),
                    float(x.get_entry(3)),
                    int(x.get_entry(4)),
                    float(x.get_entry(5))
                ) for x in conn.select(sql=query)]

                for order_tuple in order_with_payments:
                    order_id = order_tuple[0]
                    order_state = order_tuple[1]
                    order_date = order_tuple[2]

                    if order_id not in all_orders:
                        order = Order(
                            order_id=order_id,
                            date=order_date,
                            state=order_state,
                        )
                        all_orders[order_id] = order

                return list(all_orders.values())

            return self.execute_parallel_in_all_databases_returning_flat_list(
                method=inner_func,
                pos_list=self.pos_list,
            )
        except Exception as _:
            raise GetPaidOrdersException()

    def get_orders_items(
        self,
        paid_orders,    # type: List[OrderItem]
    ):
        # type: (...) -> List[OrderItem]

        try:
            def inner_func(
                conn,   # type: Connection
            ):
                # type: (...) -> List[OrderItem]

                query = self._OrderItemsQuery
                order_list = map(lambda order: order.order_id, sorted(paid_orders))
                order_list = ','.join(map(str, order_list))
                query = query.format(order_list)

                all_order_items = {}    # type: Dict[str, OrderItem]
                order_items = [(
                    int(x.get_entry(0)),
                    float(x.get_entry(1)),
                    str(x.get_entry(2)),
                    str(x.get_entry(3)),
                    str(x.get_entry(4)),
                    str(x.get_entry(5)),
                    str(x.get_entry(6))
                ) for x in conn.select(query)]

                for order_tuple in order_items:
                    prod_code = order_tuple[0]
                    quantity = order_tuple[1]
                    item_id = order_tuple[2]
                    price = order_tuple[3] if order_tuple[3] != 'None' else order_tuple[6]

                    if quantity == 0 and price == 'None':
                        continue

                    split_item_id = item_id.split('.')
                    if len(split_item_id) >= 2:
                        has_context = ".".join(split_item_id[1:]) if len(item_id) > 1 else False
                        father_context = split_item_id[1]
                    else:
                        has_context = False
                        father_context = ''

                    if father_context:
                        quantity = self._group_item_qty(
                            context=father_context,
                            order_tuple=order_tuple,
                            order_items=order_items,
                            quantity=quantity,
                        )

                    key = (str(prod_code) + '.' + has_context) if has_context else str(prod_code)
                    no_context_key = has_context and key not in all_order_items.keys()
                    no_prod_code_key = not has_context and (str(prod_code) not in all_order_items.keys())

                    if no_context_key or no_prod_code_key:
                        order_item = OrderItem(
                            name=u"",
                            quantity=quantity,
                            pcode=prod_code,
                            unit_price=price,
                            context=item_id,
                        )
                        all_order_items[key] = order_item
                    else:
                        all_order_items[key].quantity += quantity

                return list(all_order_items.values())

            return self.execute_parallel_in_all_databases_returning_flat_list(
                method=inner_func,
                pos_list=self.pos_list,
            )
        except Exception as _:
            raise GetOrderItemsException()

    @staticmethod
    def _group_item_qty(
        context,        # type: str
        order_tuple,    # type: tuple
        order_items,    # type: List
        quantity,       # type: str
    ):
        # type: (...) -> float

        order_id = order_tuple[4]
        line_number = order_tuple[5]
        father_item = filter(
            lambda item: (
                item[0] == int(context)
                and item[4] == order_id
                and line_number == item[5]
            ),
            order_items
        )
        updated_quantity = float(quantity) * father_item[0][1]
        return updated_quantity

    _PaidTicketsQuery = """
        SELECT UserId, SUM(NumberOfSeats) AS seats
        FROM TableService tb
        WHERE BusinessPeriod = '{}' 
          AND (OrderId IS NOT NULL OR TotalAmount IS NOT NULL)
        GROUP BY UserId
    """
    _PaidOrdersQuery = """
        SELECT o.OrderId, o.StateId, o.BusinessPeriod, o.TotalGross, ot.TenderId, 
            ot.TenderAmount - COALESCE(ot.ChangeAmount, 0)
        FROM Orders o
        INNER JOIN OrderTender ot
        ON o.OrderId = ot.OrderId
        WHERE o.BusinessPeriod >= '{0}' 
          AND o.BusinessPeriod <= '{1}' 
          AND o.StateId = 5 
          AND o.SessionId LIKE '%user={2},%'
    """

    _PaidOrdersTicketQuery = """\
        SELECT SessionId, COUNT(*) AS orders
        FROM orders
        WHERE StateId = 5 AND BusinessPeriod = '{}'
    """

    _OrderItemsQuery = """
        SELECT * FROM (
            SELECT PartCode, SUM(OrderedQty) AS OrderedQtySum, ItemId, DefaultUnitPrice, 
                OrderId, LineNumber, AddedUnitPrice
                FROM (
                    SELECT 
                        t.PartCode,
                        t.OrderedQty AS OrderedQty,
                        t.ItemId,
                        t.LineNumber,
                        t.OrderId,
                        p.DefaultUnitPrice,
                        p.AddedUnitPrice
                    FROM OrderItem t
                    JOIN OrderItem u ON t.OrderId = u.OrderId AND t.LineNumber = u.LineNumber
                    JOIN Orders s ON t.OrderId = s.OrderId
                    JOIN ProductKernelParams pkp ON pkp.ProductCode = t.PartCode
                    LEFT JOIN OrderPrice p 
                        ON (t.PriceKey = p.PriceKey AND t.LastOrderedQty IS NULL)
                        AND p.OrderId = t.OrderId
                    WHERE s.StateId = 5
                      AND u.Level = '0' 
                      AND t.OrderId IN ({0})
                      AND pkp.ProductType <> 1
                      AND COALESCE(t.OrderedQty, t.DefaultQty) > 0
                ) T
            GROUP BY OrderId, ItemId, PartCode, LineNumber
        ) 
        WHERE OrderedQtySum is not null
    """
